/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hama.examples;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;
import org.apache.commons.math3.linear.CholeskyDecomposition;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hama.HamaConfiguration;
import org.apache.hama.bsp.DxReadable;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.HashPartitioner2;
import org.apache.hama.bsp.MetisPartitioner;
import org.apache.hama.bsp.TextInputFormat;
import org.apache.hama.bsp.TextOutputFormat;
import org.apache.hama.graph.AverageAggregator;
import org.apache.hama.graph.DxALSErrorAggregator;
import org.apache.hama.graph.DxWasteMsgAggregator;
import org.apache.hama.graph.DxWasteNodeAggregator;
import org.apache.hama.graph.Edge;
import org.apache.hama.graph.GraphJob;
import org.apache.hama.graph.Vertex;
import org.apache.hama.graph.VertexInputReader;

public class DxALS {
	private static final Log LOG = LogFactory.getLog(DxALS.class);
	public static final double LAMBDA = 0.01;
	public static final int NLATENT = 15;
	public static final long MAX_UPDATES = Long.MAX_VALUE;
	public static double MAXIMUM_CONVERGENCE_ERROR = 0.1d;
	public static long NUSERS = 10000;
		
	//the edge data type
	public static class DxALSEdge implements Writable{

		enum EdgeRoleEnum{TRAIN, VALIDATE, PREDICT};
		private float edgeObs;
		private EdgeRoleEnum edgeRole;
		
		//default constructor
		public DxALSEdge(){
			
		}
		
		//this constructor will override the default one, need to add one manually
		public DxALSEdge(EdgeRoleEnum role, float obs){
			edgeRole = role;
			edgeObs = obs;
		}
		
		@Override
		public void write(DataOutput out) throws IOException {
			out.writeFloat(edgeObs);
			new Text(edgeRole.toString()).write(out);
		}

		@Override
		public void readFields(DataInput in) throws IOException {
			edgeObs = in.readFloat();
			Text tmp = new Text();
			tmp.readFields(in);
			edgeRole = EdgeRoleEnum.valueOf(tmp.toString());
		}
		
	}
	
	
	//since Hama requires vertexValue and Message must be the same type (M), 
	//and the message must identify where it is from, so the Message(vertexValue) type now includes vertexid(msgVid)
	//this requires that setValue should avoid reset the msgVid attribute
	public static class DxALSMessage implements Writable, DxReadable{
		
		private Text msgVid;		
		private float msgResidual;
		private int msgNupdates;
		private RealVector msgFactor;
		//used for the global error aggregation
		private double error = 0;
		
		public DxALSMessage(){
			msgVid = new Text();
			msgFactor = new ArrayRealVector(NLATENT);
		}
		
		public DxALSMessage(Text id, float res, int updates, RealVector vec, double err){
			msgVid = id;
			msgResidual = res;
			msgNupdates = updates;
			msgFactor = vec;
			error = err;
		}
		
		public DxALSMessage(DxALSMessage msg){
			msgVid = msg.msgVid;
			msgResidual = msg.msgResidual;
			msgNupdates = msg.msgNupdates;
			msgFactor = msg.msgFactor;
		}
		
		  
		@Override
		public void dxWriteBytes(DataOutput out) throws IOException {
			double[] factorArr = msgFactor.toArray();
			Assert.assertEquals(NLATENT, factorArr.length);
			for(int i = 0; i < factorArr.length; i++){			
				out.write(new Double(factorArr[i]).toString().getBytes("UTF-8"));
				out.write(" ".getBytes("UTF-8"));
			}	
		}
		
		@Override
		public void write(DataOutput out) throws IOException {
			msgVid.write(out);
			out.writeFloat(msgResidual);
			out.writeInt(msgNupdates);
			double[] factorArr = msgFactor.toArray();
			Assert.assertEquals(NLATENT, factorArr.length);
			for(int i = 0; i < factorArr.length; i++)
				out.writeDouble(factorArr[i]);
		}

		@Override
		public void readFields(DataInput in) throws IOException {
			msgVid.readFields(in);
			msgResidual = in.readFloat();
			msgNupdates = in.readInt();
			double[] factorArr = new double[NLATENT];
			for(int i = 0; i < NLATENT; i++)
				factorArr[i] = in.readDouble();
			msgFactor = new ArrayRealVector(factorArr);
		}		

		public RealMatrix getXtX(){
			Array2DRowRealMatrix XtX = new Array2DRowRealMatrix(NLATENT, NLATENT);
			
			for(int i = 0; i < NLATENT; i++)
				for(int j = 0; j < NLATENT; j++)
					XtX.setEntry(i, j, msgFactor.getEntry(i) * msgFactor.getEntry(j));
			return XtX;
		}
		
		public RealVector getXy(float obs){
			RealVector Xy = msgFactor.mapMultiply(obs);
			return Xy;
		}
		
		public RealVector getFactor(){
			return msgFactor;
		}
		
		public double getError(){
			return error;
		}
		
		public void setError(double err){
			error = err;
		}
	}
	
	public static class DxALSVertex
	extends Vertex<Text, DxALSEdge, DxALSMessage> {
		
		final class NeighborValue{
			private RealVector neighborFactor;
			private int neighborUpdates;
			
			public NeighborValue(){
				neighborFactor = new ArrayRealVector(NLATENT);
			}
			
			public NeighborValue(int num, RealVector vec){
				neighborFactor = vec;
				neighborUpdates = num;
			}
		}
		
		@Override
		public Writable getSth(){
			return new DoubleWritable(this.getValue().getError());
		}
		@Override
		public void setup(Configuration conf) {
		      String val = conf.get("hama.graph.max.convergence.error");
		      if (val != null) {
		        MAXIMUM_CONVERGENCE_ERROR = Double.parseDouble(val);
		      }
		      val = conf.get("als.nusers");
		      if(val != null){
		    	  NUSERS = Long.parseLong(val);
		      }
		    }
		
		@Override
		public void compute(Iterator<DxALSMessage> messages) throws IOException {

			if(this.getSuperstepCount() == 0){
				
				RealVector fact = new ArrayRealVector(NLATENT);
				
				Random rdm = new java.util.Random();
		        String seedStr = "";
				/*
		        if(Long.parseLong(this.getVertexID().toString()) > NUSERS)	//movie vertices [NUSERS+1, ...)
		          seedStr += (this.getNumVertices() + 1);
		        else if(Long.parseLong(this.getVertexID().toString()) <= NUSERS) //user vertices [1, NUSERS], 0 is not allowed for metis
		          seedStr += (this.getNumVertices() + 2);
		        */
				//refer to hama-0.6.0 DxALS for dataset conditions
				seedStr += this.getVertexID().toString();
		        rdm.setSeed(Long.parseLong(seedStr));
				
				for(int i = 0; i < NLATENT; i++)
					fact.setEntry(i, rdm.nextGaussian());
				
				this.setValue(new DxALSMessage(this.getVertexID(),0, 0, fact, 0));
				this.getValue().setError(0);
				
				//in the first round, only user vertices will be active
				if(Long.parseLong(this.getVertexID().toString()) > NUSERS){
					this.sendMessageToNeighbors(this.getValue());	
				}
				this.voteToHalt();
				return;
			}
			
			
			DxALSMessage value = this.getValue();
			DxALSMessage newValue = new DxALSMessage(value);
			
			
			RealMatrix XtXSum = new Array2DRowRealMatrix(NLATENT, NLATENT);
			RealVector XySum = new ArrayRealVector(NLATENT);
			
			RealVector oldFactor =  value.msgFactor;
			
			DxALSMessage msg;
			Text otherVid;
			Edge<Text, DxALSEdge> edge;
			HashMap<Text, NeighborValue> neighborMap = new HashMap<Text, NeighborValue>();
			
			while(messages.hasNext()){
				msg = messages.next();
				otherVid = msg.msgVid;
				edge = this.getEdge(otherVid);
				float obs = edge.getValue().edgeObs; //!!
				
				XtXSum = XtXSum.add(msg.getXtX());
				XySum = XySum.add(msg.getXy(obs));
				neighborMap.put(msg.msgVid, new NeighborValue(msg.msgNupdates, msg.msgFactor));	
			}									 
			
			//done with the msg processing and re-store, msgs can be deleted now to save memory
			messages = null;
			
			for(int i = 0; i < XtXSum.getRowDimension(); i++)
				XtXSum.addToEntry(i, i, LAMBDA);

			CholeskyDecomposition cd = new CholeskyDecomposition(XtXSum);
			newValue.msgFactor = cd.getSolver().solve(XySum);
			 newValue.msgResidual = (float)(newValue.msgFactor.subtract(oldFactor).getL1Norm() / XtXSum.getRowDimension()); 
			 if(newValue.msgResidual == 0){
				 LOG.info("[DX]superstep " + this.getSuperstepCount() + " vertex " + this.getVertexID() + "residual is 0");
			 }
			++newValue.msgNupdates;
			this.setValue(newValue);
			
			
			String neighborlog = "";
			
			Iterator<Edge<Text, DxALSEdge>> edgeIter = this.getEdges().iterator();
			while(edgeIter.hasNext()){
				
				edge = edgeIter.next();
				DxALSEdge edgeValue = edge.getValue();
				
				if(edgeValue.edgeRole == DxALSEdge.EdgeRoleEnum.TRAIN){
					NeighborValue neighbor = neighborMap.get(edge.getDestinationVertexID());
						RealVector otherFactor = neighbor.neighborFactor;
						int otherUpdates = neighbor.neighborUpdates;
						double predict = newValue.msgFactor.dotProduct(otherFactor);
						double prdError = Math.abs(edgeValue.edgeObs - predict);	//!edgeValue.obs
						double priority = prdError * newValue.msgResidual;
						this.getValue().setError(priority + this.getValue().getError());

					      Writable globalError = dxGetLastAggregatedValue(0);
					      if (globalError != null && this.getSuperstepCount() > 2
					          && MAXIMUM_CONVERGENCE_ERROR > ((DoubleWritable)globalError).get()) {
					        voteToHalt();
					        return;
					      }	
				}//end of if TRAIN
			}//end of while
		   this.sendMessageToNeighbors(this.getValue());  //should not send newValue
		   this.voteToHalt();
		}
		
	}
	
	public static class DxALSTextReader2 extends
    VertexInputReader<LongWritable, Text, Text, DxALSEdge, DxALSMessage> {

	  /**
	   * The text file essentially should look like: <br/>
	   * FROM_VERTEX_ID\tTO_VERTEX_ID<br/>
	   * E.G:<br/>
	   * 0\t11342<br/>
	   * 0\t824020<br/>
	   * etc.
	   */
	  String lastVertexId = null;
	  List<Edge<Text, DxALSEdge>> adjacents = new ArrayList<Edge<Text, DxALSEdge>>();
	
	  @Override
	  public boolean parseVertex(LongWritable key, Text value,
	      Vertex<Text, DxALSEdge, DxALSMessage> vertex) {
	    // the tail vertex of input
	    if (key == null || value == null) {
	      assert (lastVertexId != null);
	      vertex.setVertexID(new Text(lastVertexId));
	      for (Edge<Text, DxALSEdge> adjacent : adjacents) {
	        vertex.addEdge(adjacent);
	      }
	      adjacents.clear();
	      lastVertexId = null;
	      return true;
	    }
	
	    // normal parse
	    String line = value.toString();
	    String[] lineSplit = line.split(" ");
	    if (!line.startsWith("#")) {
	      if (lastVertexId == null) {
	        lastVertexId = lineSplit[0];
	      }
	      if (lastVertexId.equals(lineSplit[0])) {
	        adjacents.add(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
	            new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
	                .parseFloat(lineSplit[2]))));
	      } else {
	        vertex.setVertexID(new Text(lastVertexId));
	        for (Edge<Text, DxALSEdge> adjacent : adjacents) {
	          vertex.addEdge(adjacent);
	        }
	        adjacents.clear();
	        lastVertexId = lineSplit[0];
	        adjacents.add(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]),
	            new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float
	                .parseFloat(lineSplit[2]))));
	        return true;
	      }
	    }
	    return false;
	  }
	}
	
	public static class DxALSTextReader extends 
	VertexInputReader<LongWritable, Text, Text, DxALSEdge, DxALSMessage>{

		String lastVertexId = null;
	    
		@Override
		public boolean parseVertex(LongWritable key, Text value,
				Vertex<Text, DxALSEdge, DxALSMessage> vertex) throws Exception {
			return true;
		}
		
		public boolean dxParseVertex(LongWritable key, Text value, Vertex<Text, DxALSEdge, DxALSMessage> vertex,
				                     Vertex<Text, DxALSEdge, DxALSMessage> vertexNext) throws Exception
		{
			  String line = value.toString();
		      String[] lineSplit = line.split(" ");
		      
		      Assert.assertNotNull(vertexNext.getEdges());
		      Assert.assertNotNull(vertex.getEdges());
		      
		      if (!line.startsWith("#")) {
			      if (lastVertexId == null) {
			          lastVertexId = lineSplit[0];
			          vertex.setVertexID(new Text(lastVertexId));
			      }
			      if(vertexNext.getEdges().size() != 0){
			    	  Assert.assertNotNull(vertexNext.getVertexID());
	        		  vertex.setVertexID(vertexNext.getVertexID());
		        	  // only add one edge is enough for the current data input
	        		  vertex.addEdge(vertexNext.getEdges().iterator().next());
		        	  vertexNext.setVertexID(null);
		        	  vertexNext.getEdges().clear();
		          }
			      if (lastVertexId.equals(lineSplit[0])) {			        	
			        	 vertex.addEdge(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]), 
			        		  		 new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float.parseFloat(lineSplit[2]))));
			        } else {		        	
				          lastVertexId = lineSplit[0];
				          vertexNext.setVertexID(new Text(lastVertexId));
				          vertexNext.addEdge(new Edge<Text, DxALSEdge>(new Text(lineSplit[1]), 
			     		  		 new DxALSEdge(DxALSEdge.EdgeRoleEnum.TRAIN, Float.parseFloat(lineSplit[2]))));
				          
				          return true;
			        }
		      }
		     
		      return false;
		}
		
	}
	
	public static void main(String[] args) throws IOException,
    InterruptedException, ClassNotFoundException {
		HamaConfiguration conf = new HamaConfiguration(new Configuration());
	    GraphJob alsJob = createJob(args, conf);
		long startTime = System.currentTimeMillis();
	    if (alsJob.waitForCompletion(true)) {
	      System.out.println("Job Finished in "
	          + (System.currentTimeMillis() - startTime) / 1000.0 + " seconds");
	    }

	}	//end of main()
	
	public static GraphJob createJob(String[] args, HamaConfiguration conf)
		      throws IOException {
		GraphJob alsJob = new GraphJob(conf, PageRank.class);
	    alsJob.setJobName("ALS");

	    alsJob.setVertexClass(DxALSVertex.class);
	    alsJob.setInputPath(new Path(args[0]));
	    alsJob.setOutputPath(new Path(args[1]));

	    // set the defaults
	    alsJob.setPartitioner(HashPartitioner.class);
	    alsJob.setMaxIteration(60);
	    alsJob.setNumBspTask(3);
	    
	    if(args.length >= 7)
	    	alsJob.set("als.nusers", args[6]);
	    if (args.length >=6 )
	    	alsJob.set("hama.graph.max.convergence.error", args[5]);
	   
        if (args.length >= 5){
            if (args[4].equals("1"))
                alsJob.setPartitioner(MetisPartitioner.class);
        } 
	    
	    if (args.length >= 4)
	        alsJob.setMaxIteration(Integer.parseInt(args[3]));
	    if (args.length >= 3)
	        alsJob.setNumBspTask(Integer.parseInt(args[2]));	    

	    //alsJob.setAggregatorClass(AverageAggregator.class);
	    alsJob.setAggregatorClass(DxALSErrorAggregator.class);
	    
	    alsJob.setVertexIDClass(Text.class);
	    alsJob.setVertexValueClass(DxALSMessage.class);
	    alsJob.setEdgeValueClass(DxALSEdge.class);

	    alsJob.setInputKeyClass(LongWritable.class);	//input key is generated by the system
	    alsJob.setInputValueClass(Text.class);	//all your actual input is in value
	    alsJob.setInputFormat(TextInputFormat.class);
	    alsJob.setVertexInputReaderClass(DxALSTextReader2.class);
	    
	    alsJob.setOutputFormat(TextOutputFormat.class);
	    alsJob.setOutputKeyClass(Text.class);
	    alsJob.setOutputValueClass(DxALSMessage.class);
	    return alsJob;
	}
}

