/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.io.PrintWriter;
import java.io.StringWriter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.bsp.BSP;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.Combiner;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.Partitioner;
import org.apache.hama.bsp.message.queue.MessageQueue;
import org.apache.hama.bsp.sync.SyncException;
import org.apache.hama.util.KeyValuePair;

/**
 * Fully generic graph job runner.
 * 
 * @param <V> the id type of a vertex.
 * @param <E> the value type of an edge.
 * @param <M> the value type of a vertex.
 */
public final class GraphJobRunner<V extends Writable, E extends Writable, M extends Writable>
    extends BSP<Writable, Writable, Writable, Writable, GraphJobMessage> {

  public static enum GraphJobCounter {
    MULTISTEP_PARTITIONING, ITERATIONS, INPUT_VERTICES, AGGREGATE_VERTICES
  }

  private static final Log LOG = LogFactory.getLog(GraphJobRunner.class);

  // make sure that these values don't collide with the vertex names
  public static final String S_FLAG_MESSAGE_COUNTS = "hama.0";
  public static final String S_FLAG_AGGREGATOR_VALUE = "hama.1";
  public static final String S_FLAG_AGGREGATOR_INCREMENT = "hama.2";
  public static final Text FLAG_MESSAGE_COUNTS = new Text(S_FLAG_MESSAGE_COUNTS);

  public static final String MESSAGE_COMBINER_CLASS = "hama.vertex.message.combiner.class";
  public static final String GRAPH_REPAIR = "hama.graph.repair";

  private Configuration conf;
  private Combiner<M> combiner;
  private Partitioner<V, M> partitioner;

  private List<Vertex<V, E, M>> vertices = new ArrayList<Vertex<V, E, M>>();
  private Map<V, Vertex<V, E, M>> verticesMap = new HashMap<V, Vertex<V, E, M>>(); // (BSP-g) a set of master vertices

  private boolean updated = true;
  private int globalUpdateCounts = 0;
  private volatile Iterator<Entry<InetSocketAddress, MessageQueue<GraphJobMessage>>> it = null;

  private long numberVertices = 0;
  // -1 is deactivated
  private int maxIteration = -1;
  private long iteration;

  private Class<V> vertexIdClass;
  private Class<M> vertexValueClass;
  private Class<E> edgeValueClass;
  private Class<Vertex<V, E, M>> vertexClass;

  private AggregationRunner<V, E, M> aggregationRunner;

  private BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer;

  //execution time breakdown
  private long setupTOT = 0, bspTOT = 0, onceTOT = 0;
  private long setupLD = 0, setupREP = 0, setupMT = 0, setupINIT = 0;
  private long bspSYN = 0, bspSND = 0, bspPRS = 0, bspMST = 0, bspSTP = 0;
  private long onceSYN = 0, onceSND = 0, oncePRS = 0, onceMST = 0, onceSTP = 0;
  private long msgInto = 0;
  
  private long avgSetupTOT = 0, avgSetupLD = 0, avgSetupREP = 0, avgSetupREV = 0, avgSetupINIT = 0;
  private long avgBspTOT = 0, avgBspSYN = 0, avgBspSND = 0, avgBspPRS = 0, avgBspMST = 0, avgBspSTP = 0;
  private long sumMsgTOT = 0, msgSND = 0, msgRCV = 0;

  
  private MapWritable localStat = new MapWritable();
  private MapWritable globalStat = new MapWritable();
  
  @Override
  public final void setup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    
    LOG.info("Setup graph!");
    setupTOT = System.currentTimeMillis();
    
    setupFields(peer);

    loadVertices(peer);

    // convert Map to List
    this.vertices = new ArrayList<Vertex<V, E, M>>(this.verticesMap.values());
    this.verticesMap = null;

    countGlobalVertexCount(peer);

    // [SETUP-INIT] phase: initialize all vertices 
    setupINIT = System.currentTimeMillis();
    doInitialSuperstep(peer);
    it = peer.preSync();
    peer.transfer(it);
    System.gc();
    peer.postSync();
    iteration++;
    setupINIT = System.currentTimeMillis() - setupINIT;

    setupTOT = System.currentTimeMillis() - setupTOT;
    setupLD = setupTOT - setupREP - setupMT - setupINIT;
  }

  public final void breakdown() {
    LOG.info("==== setup tot: " + setupTOT + 
          " ld: " + setupLD + 
          " rep: " + setupREP + 
          " mt: " + setupMT + 
          " init: " + setupINIT + " ====");
    LOG.info("====   bsp tot: " + bspTOT + 
          " syn: " + bspSYN + 
          " prs: " + bspPRS + 
          " agg: " + bspMST + 
          " stp: " + bspSTP + 
          " snd: " + bspSND + " ====");
    if (isMasterTask(peer)) {
      LOG.info("Average");
      LOG.info("==== setup tot: " + avgSetupTOT +
          " load: " + avgSetupLD +
          " rep: " + avgSetupREP +
          " rev: " + avgSetupREV +
          " init: " + avgSetupINIT + " ====");
      LOG.info("====   bsp tot: " + avgBspTOT +
          " syn: " + avgBspSYN +
          " prs: " + avgBspPRS +
          " mst: " + avgBspMST +
          " stp: " + avgBspSTP +
          " snd: " + avgBspSND + " ====");

      LOG.info("Sum");
      LOG.info("====   msg tot: " + sumMsgTOT + " ====");
    }

  }

  @Override
  public final void bsp(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {

    LOG.info("Starting graph processing!");
    bspTOT = System.currentTimeMillis();
    
    // we do supersteps while we still have updates and have not reached our
    // maximum iterations yet
    while (updated && !((maxIteration > 0) && iteration > maxIteration)) {
      // reset the global update counter from our master in every superstep
      globalUpdateCounts = 0;

      /*
      onceSYN = System.currentTimeMillis();
      peer.sync();
      onceSYN = System.currentTimeMillis() - onceSYN;
      bspSYN += onceSYN;
      */

      // [BSP-PRS] phase: parse messages (sync-msgs and stat-msgs)
      oncePRS = System.currentTimeMillis();
      final Map<V, List<M>> messages = parseMessages(peer);
      oncePRS = System.currentTimeMillis() - oncePRS;
      bspPRS += oncePRS;
      LOG.info("Total # of msg coming into: " + msgInto);
      
      // [BSP-AGG] phase: update aggregation value if necessary (additional sync)
      onceMST = System.currentTimeMillis();
      doMasterUpdates(peer);
      // if aggregators say we don't have updates anymore, break
      if (!aggregationRunner.receiveAggregatedValues(peer, iteration)) {
        onceMST = System.currentTimeMillis() - onceMST;
        bspMST += onceMST;
        onceSTP = onceSND = 0;
        break;
      }
      onceMST = System.currentTimeMillis() - onceMST;
      bspMST += onceMST;

      // [BSP-STP] phase: loop over vertices and do their commit and computation
      onceSTP = System.currentTimeMillis();
      doSuperstep(messages, peer);
      onceSTP = System.currentTimeMillis() - onceSTP;
      bspSTP += onceSTP;
      
      // [BSP-SND] phase: transfer messages
      onceSND = System.currentTimeMillis();
      it = peer.preSync();
      peer.transfer(it);
      onceSND = System.currentTimeMillis() - onceSND;
      bspSND += onceSND;

      // [BSP-SYN] phase: sync with all workers
      onceSYN = System.currentTimeMillis();
      peer.postSync();
      onceSYN = System.currentTimeMillis() - onceSYN;
      bspSYN += onceSYN;

      onceTOT = onceSYN + oncePRS + onceMST + onceSTP + onceSND;
      LOG.info("Finish iteration#" + iteration);
      LOG.info("==== iter#" + iteration + ":" +
            " [tot] " + onceTOT + 
            " [syn] " + onceSYN + 
            " [prs] " + oncePRS +
            " [mst] " + onceMST + 
            " [stp] " + onceSTP + 
            " [snd] " + onceSND + " ====\n");
      
      this.iteration++;
      if (isMasterTask(peer)) {
        peer.getCounter(GraphJobCounter.ITERATIONS).increment(1);
      }
    }

    if (iteration > maxIteration) {
      LOG.info("BSP Job Done: max-iter (" + iteration + ")");   
    } else {
      LOG.info("BSP Job Done: converged (" + iteration + ")");
    }

    //bspTOT = System.currentTimeMillis() - bspTOT;
    bspTOT = bspSYN + bspPRS + bspMST + bspSTP + bspSND;
    average(peer);
    breakdown();

  }

  private void average(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer2) throws IOException, SyncException, InterruptedException {
    int peerIndex = peer.getPeerIndex();
    peer.sync();
    
    localStat.put(new Text("bsp-tot" + peerIndex), new LongWritable(bspTOT));
    localStat.put(new Text("bsp-syn" + peerIndex), new LongWritable(bspSYN));
    localStat.put(new Text("bsp-prs" + peerIndex), new LongWritable(bspPRS));
    localStat.put(new Text("bsp-mst" + peerIndex), new LongWritable(bspMST));
    localStat.put(new Text("bsp-stp" + peerIndex), new LongWritable(bspSTP));
    localStat.put(new Text("bsp-snd" + peerIndex), new LongWritable(bspSND));
    
    localStat.put(new Text("setup-tot" + peerIndex), new LongWritable(setupTOT));
    localStat.put(new Text("setup-ld" + peerIndex), new LongWritable(setupLD));
    localStat.put(new Text("setup-rep" + peerIndex), new LongWritable(setupREP));
    //localStat.put(new Text("setup-rev" + peerIndex), new LongWritable(setupREV));
    localStat.put(new Text("setup-init" + peerIndex), new LongWritable(setupINIT));
    
    //peer.transfer(getMasterTask(peer), localStat);
    peer.send(getMasterTask(peer), new GraphJobMessage(localStat));
    //localStat.clear();

    peer.sync();

    if (isMasterTask(peer)) {
      long btot = 0, bsyn = 0, bprs = 0, bmst = 0, bstp = 0, bsnd = 0;
      long stot = 0, sld = 0, srep = 0, srev = 0, sinit = 0;
      int numberPeers = peer.getNumPeers();

      GraphJobMessage msg = null;
      while ((msg = peer.getCurrentMessage()) != null) {
        for (Entry<Writable, Writable> e : msg.getMap().entrySet()) {
          String key = e.getKey().toString();

          if (key.startsWith("bsp-tot")) {
            btot += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("bsp-syn")) {
            bsyn += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("bsp-prs")) {
            bprs += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("bsp-mst")) {
            bmst += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("bsp-stp")) {
            bstp += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("bsp-snd")) {
            bsnd += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("setup-tot")) {
            stot += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("setup-ld")) {
            sld += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("setup-rep")) {
            srep += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("setup-rev")) {
            srev += ((LongWritable) e.getValue()).get();
          } else if (key.startsWith("setup-init")) {
            sinit += ((LongWritable) e.getValue()).get();
          } else {
            throw new UnsupportedOperationException("unknown time: " + key);
          }
        }
      }

      avgBspTOT = btot / numberPeers;
      avgBspSYN = bsyn / numberPeers;
      avgBspPRS = bprs / numberPeers;
      avgBspMST = bmst / numberPeers;
      avgBspSTP = bstp / numberPeers;
      avgBspSND = bsnd / numberPeers;

      avgSetupTOT = stot / numberPeers;
      avgSetupLD = sld / numberPeers;
      avgSetupREP = srep / numberPeers;
      avgSetupREV = srev / numberPeers;
      avgSetupINIT = sinit / numberPeers;
    }


    
  }

  /**
   * Just write <ID as Writable, Value as Writable> pair as a result. Note that
   * this will also be executed when failure happened.
   */
  @Override
  public final void cleanup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    for (Vertex<V, E, M> e : vertices) {
      peer.write(e.getVertexID(), e.getValue());
    }
  }

  /**
   * The master task is going to check the number of updated vertices and do
   * master aggregation. In case of no aggregators defined, we save a sync by
   * reading multiple typed messages.
   */
  private void doMasterUpdates(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    if (isMasterTask(peer) && iteration > 1) {
      MapWritable updatedCnt = new MapWritable();
      // exit if there's no update made
      if (globalUpdateCounts == 0) {
        updatedCnt.put(FLAG_MESSAGE_COUNTS, new IntWritable(Integer.MIN_VALUE));
      } else {
        aggregationRunner.doMasterAggregation(updatedCnt);
      }
      // send the updates from the mater tasks back to the slaves
      for (String peerName : peer.getAllPeerNames()) {
        peer.send(peerName, new GraphJobMessage(updatedCnt));
      }
    }
  }

  /**
   * Do the main logic of a superstep, namely checking if vertices are active,
   * feeding compute with messages and controlling combiners/aggregators.
   */
  private void doSuperstep(Map<V, List<M>> messages,
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    int activeVertices = 0;
    for (Vertex<V, E, M> vertex : vertices) {
      List<M> msgs = messages.get(vertex.getVertexID());
      // If there are newly received messages, restart.
      if (vertex.isHalted() && msgs != null) {
        vertex.setActive();
      }
      if (msgs == null) {
        msgs = Collections.emptyList();
      }

      if (!vertex.isHalted()) {
    	activeVertices++;
        if (combiner != null) {
          M combined = combiner.combine(msgs);
          msgs = new ArrayList<M>();
          msgs.add(combined);
        }
        M lastValue = vertex.getValue();
        vertex.compute(msgs.iterator());
        aggregationRunner.aggregateVertex(lastValue, vertex);
        /*if (!vertex.isHalted()) {
          activeVertices++;
        }*/
      }
    }

    aggregationRunner.sendAggregatorValues(peer, activeVertices);
  }

  /**
   * Seed the vertices first with their own values in compute. This is the first
   * superstep after the vertices have been loaded.
   */
  private void doInitialSuperstep(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    //long edgeNumber=0;
    for (Vertex<V, E, M> vertex : vertices) {
      //edgeNumber += vertex.getEdges().size();
      List<M> singletonList = Collections.singletonList(vertex.getValue());
      M lastValue = vertex.getValue();
      vertex.compute(singletonList.iterator());
      aggregationRunner.aggregateVertex(lastValue, vertex);
    }

    aggregationRunner.sendAggregatorValues(peer, 1);
    //LOG.info("Total # of out-edge: " + edgeNumber);
  }

  @SuppressWarnings("unchecked")
  private void setupFields(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    this.peer = peer;
    this.conf = peer.getConfiguration();
    maxIteration = peer.getConfiguration().getInt("hama.graph.max.iteration",
        -1);

    vertexIdClass = (Class<V>) conf.getClass(GraphJob.VERTEX_ID_CLASS_ATTR,
        Text.class, Writable.class);
    vertexValueClass = (Class<M>) conf.getClass(
        GraphJob.VERTEX_VALUE_CLASS_ATTR, IntWritable.class, Writable.class);
    edgeValueClass = (Class<E>) conf.getClass(
        GraphJob.VERTEX_EDGE_VALUE_CLASS_ATTR, IntWritable.class,
        Writable.class);
    vertexClass = (Class<Vertex<V, E, M>>) conf.getClass(
        "hama.graph.vertex.class", Vertex.class);

    // set the classes statically, so we can save memory per message
    GraphJobMessage.VERTEX_ID_CLASS = vertexIdClass;
    GraphJobMessage.VERTEX_VALUE_CLASS = vertexValueClass;
    GraphJobMessage.VERTEX_CLASS = vertexClass;
    GraphJobMessage.EDGE_VALUE_CLASS = edgeValueClass;

    partitioner = (Partitioner<V, M>) ReflectionUtils.newInstance(
        conf.getClass("bsp.input.partitioner.class", HashPartitioner.class),
        conf);
    partitioner.setConf(conf);

    if (!conf.getClass(MESSAGE_COMBINER_CLASS, Combiner.class).equals(
        Combiner.class)) {
      LOG.debug("Combiner class: " + conf.get(MESSAGE_COMBINER_CLASS));

      combiner = (Combiner<M>) ReflectionUtils.newInstance(
          conf.getClass("hama.vertex.message.combiner.class", Combiner.class),
          conf);
    }

    aggregationRunner = new AggregationRunner<V, E, M>();
    aggregationRunner.setupAggregators(peer);
  }

  /**
   * Loads vertices into memory of each peer. TODO this needs to be simplified.
   */
  @SuppressWarnings("unchecked")
  private void loadVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {

    /*
     * Several partitioning constants begin
     */

    final VertexInputReader<Writable, Writable, V, E, M> reader = 
        (VertexInputReader<Writable, Writable, V, E, M>) ReflectionUtils
        .newInstance(conf.getClass(GraphJob.VERTEX_GRAPH_INPUT_READER,
            VertexInputReader.class), conf);

    final boolean repairNeeded = conf.getBoolean(GRAPH_REPAIR, false);
    final boolean runtimePartitioning = conf.getBoolean(
        GraphJob.VERTEX_GRAPH_RUNTIME_PARTIONING, true);

    final long splitSize = peer.getSplitSize();
    final int partitioningSteps = partitionMultiSteps(peer, splitSize);
    final long interval = splitSize / partitioningSteps;

    final boolean selfReference = conf.getBoolean("hama.graph.self.ref", false);

    /*
     * Several partitioning constants end
     */

    Vertex<V, E, M> vertex = newVertexInstance(vertexClass, conf);
    vertex.runner = this;

    long startPos = peer.getPos();
    if (startPos == 0)
      startPos = 1L;

    KeyValuePair<Writable, Writable> next = null;
    int steps = 1;
    boolean pending = false; // (BSP-g) fix the tail vertex bug
    while ((next = peer.readNext()) != null) {
      boolean vertexFinished = false;
      try {
        vertexFinished = reader.parseVertex(next.getKey(), next.getValue(),
            vertex);
        pending = true; // (BSP-g)
      } catch (Exception e) {
        LOG.error("exception occurred during parsing vertex!" + e.toString());
        StringWriter errors = new StringWriter();
	e.printStackTrace(new PrintWriter(errors));
        LOG.error(errors.toString());
        throw new IOException("exception occured during parsing vertex!"
            + e.toString());
      }

      if (!vertexFinished) {
        continue;
      }

      if (vertex.getEdges() == null) {
        if (selfReference) {
          vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
              .getVertexID(), null)));
        } else {
          vertex.setEdges(Collections.EMPTY_LIST);
        }
      }

      if (selfReference) {
        vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));
      }

      if (runtimePartitioning) {
        int partition = partitioner.getPartition(vertex.getVertexID(),
            vertex.getValue(), peer.getNumPeers());
        peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
      } else {
        vertex.setup(conf);
        verticesMap.put(vertex.getVertexID(), vertex); // (BSP-g) List -> Map
      }
      vertex = newVertexInstance(vertexClass, conf);
      vertex.runner = this;

      if (runtimePartitioning) {
        if (steps < partitioningSteps && (peer.getPos() - startPos) >= interval) {
          peer.sync();
          steps++;
          collect(peer);
          startPos = peer.getPos();
        }
      }
    }

    // (BSP-g) fix the tail vertex bug
    if (pending) {
      boolean done = false;
      try {
        done = reader.parseVertex(null, null, vertex);
      } catch (Exception e) {
        throw new IOException("exception occured during parsing tail vertex!"
            + e.toString());
      }
      if (done) {
        if (vertex.getEdges() == null) {
          if (selfReference) {
            vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
                .getVertexID(), null)));
          } else {
            vertex.setEdges(Collections.EMPTY_LIST);
          }
        }

        if (selfReference) {
          vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));
        }
      
        if (runtimePartitioning) {
          int partition = partitioner.getPartition(vertex.getVertexID(),
              vertex.getValue(), peer.getNumPeers());
          peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
        } else {
          vertex.setup(conf);
          verticesMap.put(vertex.getVertexID(), vertex); // (BSP-g) List -> Map
        }
        LOG.info("Fix the tail vertex: " + vertex.getVertexID());
      }
    }

    // the last step
    if (runtimePartitioning) {
      peer.sync();
      collect(peer);
    }

    LOG.info("Loading finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());

    /*
     * If the user want to repair the graph, it should traverse through that
     * local chunk of adjancency list and message the corresponding peer to
     * check whether that vertex exists. In real-life this may be dead-ending
     * vertices, since we have no information about outgoing edges. Mainly this
     * procedure is to prevent NullPointerExceptions from happening.
     */
    if (repairNeeded) {
      LOG.info("Starting repair of this graph!");
      repair(peer, partitioningSteps, selfReference);
      LOG.info("Repair finished at " + peer.getSuperstepCount() + " steps.");
      LOG.info("Total # of local vertices: " + verticesMap.size());
    }
  }

  @SuppressWarnings("unchecked")
  private void collect(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException {
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      Vertex<V, E, M> vertex = (Vertex<V, E, M>) msg.getVertex();
      vertex.runner = this;
      
      Vertex<V, E, M> old = verticesMap.put(vertex.getVertexID(), vertex);
      if (old != null)
        vertex.merge(old);

      // call setup after merge
      vertex.setup(conf);
    }
  }

  @SuppressWarnings("unchecked")
  private void repair(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      int partitioningSteps, boolean selfReference) throws IOException,
      SyncException, InterruptedException {

    int multiSteps = 0;
    MapWritable ssize = new MapWritable();
    ssize.put(new IntWritable(peer.getPeerIndex()),
        new IntWritable(verticesMap.size()));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      int minVerticesSize = Integer.MAX_VALUE;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          int curr = ((IntWritable) e.getValue()).get();
          if (minVerticesSize > curr) {
            minVerticesSize = curr;
          }
        }
      }

      if (minVerticesSize < (partitioningSteps * 2)) {
        multiSteps = minVerticesSize;
      } else {
        multiSteps = (partitioningSteps * 2);
      }

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("steps"), new IntWritable(multiSteps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    Map<V, Vertex<V, E, M>> tmp = new HashMap<V, Vertex<V, E, M>>();

    int i = 0;
    int syncs = 0;

    for (Vertex<V, E, M> v : verticesMap.values()) {
      for (Edge<V, E> e : v.getEdges()) {
        peer.send(v.getDestinationPeerName(e),
            new GraphJobMessage(e.getDestinationVertexID()));
      }

      if (syncs < multiSteps && (i % (verticesMap.size() / multiSteps)) == 0) {
        peer.sync();
        syncs++;
        GraphJobMessage msg = null;
        while ((msg = peer.getCurrentMessage()) != null) {
          V vertexName = (V) msg.getVertexId();

          Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
          newVertex.setVertexID(vertexName);
          newVertex.runner = this;
          if (selfReference) {
            newVertex.setEdges(Collections.singletonList(new Edge<V, E>(
                newVertex.getVertexID(), null)));
          } else {
            newVertex.setEdges(new ArrayList<Edge<V, E>>(0));
          }
          newVertex.setup(conf);
          tmp.put(vertexName, newVertex);

        }
      }
      i++;
    }

    peer.sync();
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      V vertexName = (V) msg.getVertexId();

      Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
      newVertex.setVertexID(vertexName);
      newVertex.runner = this;
      if (selfReference) {
        newVertex.setEdges(Collections.singletonList(new Edge<V, E>(newVertex
            .getVertexID(), null)));
      } else {
        newVertex.setEdges(new ArrayList<Edge<V, E>>(0));
      }
      newVertex.setup(conf);
      tmp.put(vertexName, newVertex);
      newVertex = null;

    }

    for (Vertex<V, E, M> e : verticesMap.values()) {
      if (tmp.containsKey((e.getVertexID()))) {
        tmp.remove(e.getVertexID());
      }
    }

    verticesMap.putAll(tmp);
    tmp.clear();
  }

  /**
   * Partitions our vertices through multiple supersteps to save memory.
   */
  private int partitionMultiSteps(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      long splitSize) throws IOException, SyncException, InterruptedException {
    int multiSteps = 1;

    MapWritable ssize = new MapWritable();
    ssize
        .put(new IntWritable(peer.getPeerIndex()), new LongWritable(splitSize));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      long maxSplitSize = 0L;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          long curr = ((LongWritable) e.getValue()).get();
          if (maxSplitSize < curr) {
            maxSplitSize = curr;
          }
        }
      }

      int steps = (int) (maxSplitSize / conf.getLong( // 20 mb
          "hama.graph.multi.step.partitioning.interval", 20000000)) + 1;

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("max"), new IntWritable(steps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    if (isMasterTask(peer)) {
      peer.getCounter(GraphJobCounter.MULTISTEP_PARTITIONING).increment(
          multiSteps);
    }

    return multiSteps;
  }

  /**
   * Counts vertices globally by sending the count of vertices in the map to the
   * other peers.
   */
  private void countGlobalVertexCount(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    for (String peerName : peer.getAllPeerNames()) {
      peer.send(peerName, new GraphJobMessage(new IntWritable(vertices.size())));
    }

    peer.sync();

    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      if (msg.isVerticesSizeMessage()) {
        numberVertices += msg.getVerticesSize().get();
      }
    }

    if (isMasterTask(peer)) {
      peer.getCounter(GraphJobCounter.INPUT_VERTICES).increment(numberVertices);
    }
    
    LOG.info("Counting finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of vertices: " + numberVertices);
  }

  /**
   * Parses the messages in every superstep and does actions according to flags
   * in the messages.
   * 
   * @return a map that contains messages pro vertex.
   */
  @SuppressWarnings("unchecked")
  private Map<V, List<M>> parseMessages(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    GraphJobMessage msg = null;
    final Map<V, List<M>> msgMap = new HashMap<V, List<M>>();

    LOG.info("Parsing #message = " + peer.getNumCurrentMessages());
    while ((msg = peer.getCurrentMessage()) != null) {
      // either this is a vertex message or a directive that must be read
      // as map
      if (msg.isVertexMessage()) {
        final V vertexID = (V) msg.getVertexId();
        final M value = (M) msg.getVertexValue();
        List<M> msgs = msgMap.get(vertexID);
        if (msgs == null) {
          msgs = new ArrayList<M>();
          msgMap.put(vertexID, msgs);
        }
        msgs.add(value);
      } else if (msg.isMapMessage()) {
        for (Entry<Writable, Writable> e : msg.getMap().entrySet()) {
          Text vertexID = (Text) e.getKey();
          if (FLAG_MESSAGE_COUNTS.equals(vertexID)) {
            if (((IntWritable) e.getValue()).get() == Integer.MIN_VALUE) {
              updated = false;
            } else {
              globalUpdateCounts += ((IntWritable) e.getValue()).get();
            }
          } else if (aggregationRunner.isEnabled()
              && vertexID.toString().startsWith(S_FLAG_AGGREGATOR_VALUE)) {
            aggregationRunner.masterReadAggregatedValue(vertexID,
                (M) e.getValue());
          } else if (aggregationRunner.isEnabled()
              && vertexID.toString().startsWith(S_FLAG_AGGREGATOR_INCREMENT)) {
            aggregationRunner.masterReadAggregatedIncrementalValue(vertexID,
                (M) e.getValue());
          }
        }
      } else {
        throw new UnsupportedOperationException("Unknown message type: " + msg);
      }

    }
    return msgMap;
  }

  /**
   * @return the number of vertices, globally accumulated.
   */
  public final long getNumberVertices() {
    return numberVertices;
  }

  /**
   * @return the current number of iterations.
   */
  public final long getNumberIterations() {
    return iteration;
  }

  /**
   * @return the defined number of maximum iterations, -1 if not defined.
   */
  public final int getMaxIteration() {
    return maxIteration;
  }

  /**
   * @return the defined partitioner instance.
   */
  public final Partitioner<V, M> getPartitioner() {
    return partitioner;
  }

  /**
   * Gets the last aggregated value at the given index. The index is dependend
   * on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final Writable getLastAggregatedValue(int index) {
    return aggregationRunner.getLastAggregatedValue(index);
  }

  /**
   * Gets the last aggregated number of vertices at the given index. The index
   * is dependend on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final IntWritable getNumLastAggregatedVertices(int index) {
    return aggregationRunner.getNumLastAggregatedVertices(index);
  }

  /**
   * @return the peer instance.
   */
  public final BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> getPeer() {
    return peer;
  }

  /**
   * Checks if this is a master task. The master task is the first peer in the
   * peer array.
   */
  public static boolean isMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName().equals(getMasterTask(peer));
  }

  /**
   * @return the name of the master peer, the name at the first index of the
   *         peers.
   */
  public static String getMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName(0);
  }

  /**
   * @return a new vertex instance
   */
  public static <V extends Writable, E extends Writable, M extends Writable> Vertex<V, E, M> newVertexInstance(
      Class<?> vertexClass, Configuration conf) {
    @SuppressWarnings("unchecked")
    Vertex<V, E, M> vertex = (Vertex<V, E, M>) ReflectionUtils.newInstance(
        vertexClass, conf);
    return vertex;
  }

}
