/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.bsp;

import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.Writable;

public abstract class AbstractVertex<V extends Writable, E extends Writable, M extends Writable> {
  
  public static enum VertexType {
    MASTER, REPLICA, GHOST
  }
  
  protected VertexType vertexType;
  protected V vertexID;
  
  protected M value;
  //TODO: we suppose the message value used by neighbors are the same, maybe we could design a meta-data
  protected M message;

  //the value and message set by last iteration, and used by current iteration
  protected M lastValue; 
  protected M lastMessage;
  protected boolean updated = false;

  // out-edges in local peer
  protected List<AbstractRefEdge<V, E, M>> outEdges = new ArrayList<AbstractRefEdge<V, E, M>>() ;
  protected boolean awaked = false;

  public void setVertexType(VertexType type) {
    this.vertexType = type;
  }
  
  public VertexType getVertexType() {
    return this.vertexType;
  }
  
  public boolean isMaster() {
    return this.vertexType == VertexType.MASTER;
  }
  
  public V getVertexID() {
    return vertexID;
  }
  
  public void setVertexID(V vertexID) {
    this.vertexID = vertexID;
  }
  
  public M getLastValue() {
    return this.lastValue;
  }
  
  public M getNewValue() {
    return this.value;
  }
  
  public void setLastValue(Writable value) {
    this.lastValue = (M) value;
  }

  public void setNewValue(Writable value) {
    this.value = (M) value;
    this.updated = true;
  }

  public M getLastMessage() {
    return this.lastMessage;
  }

  public M getNewMessage() {
    return this.message;
  }
  
  public void setLastMessage(Writable message) {
    this.lastMessage = (M) message;
  }

  public void setNewMessage(Writable message) {
    this.message = (M) message;
    this.updated = true;
  }

  public void addOutEdge(AbstractRefEdge<V, E, M> edge) {
    this.outEdges.add(edge);
  }

  @SuppressWarnings("unchecked")
  public long awake(List<Writable> list) {
    int num = 0;

    for (Writable id : list) {
      for (AbstractRefEdge<V, E, M> e : this.outEdges) {
        if (e.getVertexID().equals((V) id)) {
          AbstractVertex<V, E, M> neighbor = e.getVertex();
          neighbor.awaked = true;
          e.getPair().activate();
          num++;
          break;
        }
      }
    }
    return num;
  }
  
  /**
   * (BSP-g) awake neighbors
   */  
  public long awakeNeighbors() {
    for (AbstractRefEdge<V, E, M> e: this.outEdges) {
      AbstractVertex<V, E, M> neighbor = e.getVertex();
      neighbor.awaked = true;
      e.getPair().activate();
    }
    return this.outEdges.size();
  }
}
