/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.hama.graph;

import java.io.IOException;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import junit.framework.Assert;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.BooleanWritable;
import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.MapWritable;
import org.apache.hadoop.io.NullWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.Writable;
import org.apache.hadoop.util.ReflectionUtils;
import org.apache.hama.bsp.AbstractVertex.VertexType;
import org.apache.hama.bsp.BSP;
import org.apache.hama.bsp.BSPPeer;
import org.apache.hama.bsp.Combiner;
import org.apache.hama.bsp.HashPartitioner;
import org.apache.hama.bsp.LyraMessageBundle;
import org.apache.hama.bsp.LyraSyncMessage;
import org.apache.hama.bsp.LyraTestMessage;
import org.apache.hama.bsp.LyraTestMessageBundle;
import org.apache.hama.bsp.Partitioner;
import org.apache.hama.bsp.sync.SyncException;
import org.apache.hama.graph.RefEdge.RefEdgeType;
import org.apache.hama.graph.Vertex;
import org.apache.hama.util.KeyValuePair;

/**
 * Fully generic graph job runner.
 * 
 * @param <V> the id type of a vertex.
 * @param <E> the value type of an edge.
 * @param <M> the value type of a vertex.
 */
public final class GraphJobRunner<V extends Writable, E extends Writable, M extends Writable>
    extends BSP<Writable, Writable, Writable, Writable, GraphJobMessage> {

  public static enum GraphJobCounter {
    MULTISTEP_PARTITIONING, ITERATIONS, INPUT_VERTICES, AGGREGATE_VERTICES
  }

  private static final Log LOG = LogFactory.getLog(GraphJobRunner.class);
  
  // make sure that these values don't collide with the vertex names
  public static final String S_FLAG_STAT = "hama.stat";
  public static final String S_FLAG_HALT = "hama.stat.0";
  public static final String S_FLAG_ALIVE_COUNTS = "hama.stat.1";
  public static final String S_FLAG_LOCAL_MESSAGE_COUNTS = "hama.stat.2";
  public static final String S_FLAG_REMOTE_MESSAGE_COUNTS = "hama.stat.3";
  public static final String S_FLAG_LAST_ACTIVE_COUNTS = "hama.stat.4";
  public static final String S_FLAG_SYNC_MESSAGE_COUNTS = "hama.stat.5";
  
  public static Text FLAG_STAT_HALT;
  public static Text FLAG_STAT_ALIVE_COUNTS;   
  public static Text FLAG_STAT_LOCAL_MESSAGE_COUNTS;
  public static Text FLAG_STAT_REMOTE_MESSAGE_COUNTS;
  public static Text FLAG_STAT_LAST_ACTIVE_COUNTS;
  public static Text FLAG_STAT_SYNC_MESSAGE_COUNTS;
  
  public static final String S_FLAG_AGGREGATOR_VALUE = "hama.1";
  public static final String S_FLAG_AGGREGATOR_INCREMENT = "hama.2";
  
  public static final String MESSAGE_COMBINER_CLASS = "hama.vertex.message.combiner.class";
  public static final String GRAPH_REPAIR = "hama.graph.repair";
  public static final String GRAPH_SELECTIVE_ACTIVATE = "hama.graph.selective.activate";

  private Configuration conf;
  private Combiner<M> combiner;
  private Partitioner<V, M> partitioner;
  private int multiSteps = 1;

  private List<Vertex<V, E, M>> vertices;
  private Map<V, Vertex<V, E, M>> verticesMap = new HashMap<V, Vertex<V, E, M>>();
  private Vertex<V, E, M>[] replicas;
  private Map<V, Vertex<V, E, M>> replicasMap = new HashMap<V, Vertex<V, E, M>>();
  
  private LyraMessageBundle[] bundles;
  private int bundleSize = 5000;
  // stat
  private long globalActiveVertexCount = 0;
  private long globalAliveVertexCount = 0;
  private long globalLocalMessageCount = 0;
  private long globalRemoteMessageCount = 0;
  private long globalSyncMessageCount = 0;
  private MapWritable globalStat = new MapWritable();
  
  private long localActiveVertexCount = 0;
  private long localAliveVertexCount = 0;
  private long localLocalMessageCount = 0;
  private long localRemoteMessageCount = 0;
  private long localSyncMessageCount = 0;
  private MapWritable localStat = new MapWritable();
  
  private long numberVertices = 0;
  private long numberReplicas = 0;
  // -1 is deactivated
  private int maxIteration = -1;
  private long iteration;
  private float minRate = -1f;  
  private float rate;  
  private volatile boolean converged = true;

  private Class<V> vertexIdClass;
  private Class<M> vertexValueClass;
  private Class<E> edgeValueClass;
  private Class<Vertex<V, E, M>> vertexClass;

  private AggregationRunner<V, E, M> aggregationRunner;

  private BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer;

  private boolean selectiveActivate = false;
  private boolean runtimePartitioning = false;
  
  // multi-thread support
  private boolean multiThreads = false;
  private int numberThreads = 2;
  private int numberReceivers = 2;

  private volatile ThreadPoolExecutor threadPool;
  private ExecutorCompletionService<Integer> completionService;
  private final Lock runnerLock = new ReentrantLock();
  private final Condition subRunnerCondition = runnerLock.newCondition();
  private final Condition runnerCondition = runnerLock.newCondition();
  private final AtomicInteger waiters = new AtomicInteger(0);
  
  private volatile long subActiveVertexCounts[] = null;
  private volatile long subAliveVertexCounts[] = null;
  private volatile long subLocalMessageCounts[] = null;
  private volatile long subRemoteMessageCounts[] = null;

  //execution time breakdown
  private long setupTOT = 0, bspTOT = 0, onceTOT = 0;
  private long setupLD = 0, setupREP = 0, setupREV = 0, setupINIT = 0;
  private long bspSYN = 0, bspSND = 0, bspPRS = 0, bspMST = 0, bspSTP = 0;
  private long onceSYN = 0, onceSYN2 = 0, onceSND = 0, onceMST = 0, onceSTP = 0;
  private long avgSetupTOT = 0, avgSetupLD = 0, avgSetupREP = 0, avgSetupREV = 0, avgSetupINIT = 0;
  private long avgBspTOT = 0, avgBspSYN = 0, avgBspSND = 0, avgBspPRS = 0, avgBspMST = 0, avgBspSTP = 0;
  private long sumMsgTOT = 0, msgSND = 0, msgRCV = 0;

  //this class will spawn a new thread and executes a subtask
  class GraphJobSubRunner implements Callable<Integer> {
    
    private final int id;
    private BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer;
    private CyclicBarrier barrier;
    private int subIteration = 0;

    private List<Vertex<V, E, M>> subVertices;
    private LyraMessageBundle[] subBundles;
    
    public GraphJobSubRunner(BSPPeer<Writable, Writable, Writable, Writable,
        GraphJobMessage> peer, int id, List<Vertex<V, E, M>> vertices, 
        CyclicBarrier barrier) {
      super();
      this.peer = peer;
      this.id = id;
      this.subVertices = vertices;
      this.barrier = barrier;
      
      subBundles = new LyraMessageBundle[peer.getNumPeers()];
      for (int i = 0; i < subBundles.length; i++)
        subBundles[i] = new LyraMessageBundle();
      peer.registerMessaging(id, replicas, globalStat);
    }
    
    public void localSync() throws SyncException {
      try {
        this.barrier.await();
      } catch (Exception e) {
        throw new SyncException(e.toString());
      }
    }
    
    public void globalSync() throws SyncException {
      runnerLock.lock();
      try {
        if (numberThreads == waiters.incrementAndGet()) {
          runnerCondition.signalAll();
          //LOG.info("sub-runner after singalALL " + waiters.get());
          if (waiters.get() != 0) {
            //LOG.info("sub-runner begin await " + waiters.get());
            subRunnerCondition.await();
            //LOG.info("sub-runner after await " + waiters.get());
          }
        } else {
          //LOG.info("sub-runner begin await XX " + waiters.get());
          subRunnerCondition.await();
          //LOG.info("sub-runner after await XX " + waiters.get());
        }
      } catch (Exception e) {
        throw new SyncException(e.toString());
      } finally {
        runnerLock.unlock();
      }
    }
    
    public void globalSyncFinit() throws SyncException {
      runnerLock.lock();
      try {
        if (numberThreads == waiters.incrementAndGet()) {
          runnerCondition.signalAll();
        }
      } catch (Exception e) {
        throw new SyncException(e.toString());
      } finally {
        runnerLock.unlock();
      }
    }

    private void init() throws IOException {
      long activeVertexCounts = 0;
      long aliveVertexCounts = 0;
      long localMessageCounts = 0;
      long remoteMessageCounts = 0;

      //LOG.info("sub-runner#" + this.id + ": start do sub superstep. | iter[" + this.subIteration + "]");
      for (Vertex<V, E, M> vertex : this.subVertices) {
        activeVertexCounts++;
        
        //M lastValue = vertex.getLastValue();
        //TODO: modify compute function interface?
        vertex.compute(null);
        localMessageCounts += vertex.localSync();
        remoteMessageCounts += vertex.remoteSync(subBundles);
        
        if (!vertex.isConverged()) {
          aliveVertexCounts++;
        }
      }

      // [unused] send aggregation value if necessary (need additional sync)
      //aggregationRunner.sendAggregatorValues(peer);
      
      // collect local stat of sub-runner and send to master runner
      subActiveVertexCounts[id] = activeVertexCounts;
      subAliveVertexCounts[id] = aliveVertexCounts;
      subLocalMessageCounts[id] = localMessageCounts;
      subRemoteMessageCounts[id] = remoteMessageCounts;
      
      /*LOG.info("sub-runner#" + this.id + ": done. | iter[" + this.subIteration + "] "
          + " # of active: " + activeVertices 
          + " local-msgs: " + localMessages
          + " remote-msgs: " + remoteMessages
          + " alive: " + aliveVertices);*/
      
      this.subIteration++;
    }
    
    private void commit() throws IOException {
      for (Vertex<V, E, M> vertex : this.subVertices)
        vertex.commit();
    }

    private void compute() throws IOException {
      long activeVertexCounts = 0;
      long aliveVertexCounts = 0;
      long localMessageCounts = 0;
      long remoteMessageCounts = 0;

      //LOG.info("sub-runner#" + this.id + ": start do sub superstep. | iter[" + this.subIteration + "]");
      for (Vertex<V, E, M> vertex : this.subVertices) {
        if (!vertex.isHalted()) {
          activeVertexCounts++;
          
          //M lastValue = vertex.getLastValue();
          //TODO: modify compute function interface?
          vertex.compute(null);
          localMessageCounts += vertex.localSync();
          remoteMessageCounts += vertex.remoteSync(subBundles);
          
          //aggregationRunner.aggregateVertex(lastValue, vertex);
          
          if (!vertex.isConverged())
            aliveVertexCounts++;
        }          
      }
      
      // [unused] send aggregation value if necessary (need additional sync)
      //aggregationRunner.sendAggregatorValues(peer);
      
      // collect local stat of sub-runner and send to master runner
      subActiveVertexCounts[id] = activeVertexCounts;
      subAliveVertexCounts[id] = aliveVertexCounts;
      subLocalMessageCounts[id] = localMessageCounts;
      subRemoteMessageCounts[id] = remoteMessageCounts;
      
      /*LOG.info("sub-runner#" + this.id + ": done. | iter[" + this.subIteration + "] "
          + " # of active: " + activeVertices 
          + " local-msgs: " + localMessages
          + " remote-msgs: " + remoteMessages
          + " alive: " + aliveVertices);*/
      
      this.subIteration++;
    }
 
    private void send() {
      // send sync-msgs
      /*for (int i = 0; i < subBundles.length; i++) {
        if (!subBundles[i].isEmpty()) {
          peer.transfer(id, peer.getPeerName(i), subBundles[i]);
          subBundles[i] = new LyraMessageBundle();
        }
      }*/
      
      for (int i = 0; i < subBundles.length; i++) {
        if (!subBundles[i].isEmpty()) {
          int n = subBundles[i].size() / bundleSize;
          if (n > 1) {
            List<LyraMessageBundle> list = subBundles[i].split(n);
            for (LyraMessageBundle bundle: list)
              peer.transfer(id, peer.getPeerName(i), bundle);
          } else {
            peer.transfer(id, peer.getPeerName(i), subBundles[i]);
          }
          
          subBundles[i] = new LyraMessageBundle();
        }
      }

      
      // send local stat
      if (this.isCoordinator()) {
        for (int i = 0; i < numberThreads; i++) {
          localActiveVertexCount += subActiveVertexCounts[i];
          localAliveVertexCount += subAliveVertexCounts[i];
          localLocalMessageCount += subLocalMessageCounts[i];
          localRemoteMessageCount += subRemoteMessageCounts[i];
        }

        if ((localAliveVertexCount>0) || (localLocalMessageCount>0) || (localRemoteMessageCount>0))
          localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(false));
        else 
          localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(true));
        localStat.put(GraphJobRunner.FLAG_STAT_LAST_ACTIVE_COUNTS, new LongWritable(localActiveVertexCount));
        localStat.put(GraphJobRunner.FLAG_STAT_ALIVE_COUNTS, new LongWritable(localAliveVertexCount));
        localStat.put(GraphJobRunner.FLAG_STAT_LOCAL_MESSAGE_COUNTS, new LongWritable(localLocalMessageCount));
        localStat.put(GraphJobRunner.FLAG_STAT_REMOTE_MESSAGE_COUNTS, new LongWritable(localRemoteMessageCount));
        // calculate in arbitrate() by main thread
        localStat.put(GraphJobRunner.FLAG_STAT_SYNC_MESSAGE_COUNTS, new LongWritable(localSyncMessageCount));
        
        for (String peerName : peer.getAllPeerNames())
          peer.transfer(id, peerName, localStat);
        localStat.clear();
        
        /*LOG.info("worker:"
            + " [msg] #local = " + localLocalMessageCount
            + " #remote = " + localRemoteMessageCount
            + " [vtx] #active: " + localActiveVertexCount
            + " #alive: " + localAliveVertexCount
            + " [rep] #synced: " + localSyncMessageCount);*/
      }
    }
    
    
    public boolean isCoordinator() {
      return this.id == 0;
    }
    
    @Override
    public Integer call() throws Exception {
      try {
        //LOG.info("sub-runner#" + this.id + ": iter[" + this.subIteration + "] | start.");
        
        // [SETUP-Init]: initialize vertices
        commit();
        localSync();        
        init();
        peer.prepare(id);
        
        globalSync(); // [BSP-SYN1]
        send();
        globalSync(); // [BSP-SYN2]
        
        while (!isDone()) {
          if (isConverged())
            break; // done
          
          // [BSP-CMP]: loop over vertices and do commit and computation
          commit();
          localSync();
          compute();
          peer.prepare(id);
          
          globalSync(); // [BSP-SYN1]
          send(); // [BSP-SND]: send sync and stat msgs
          globalSync(); // [BSP-SYN2]
        }
      } catch (Exception e) {
        LOG.info("sub-runner#" + this.id + ": iter[" + this.subIteration + 
                "] | catch fatal exception: " + e);
        e.printStackTrace();
        System.exit(1);
      }
      
      //LOG.info("sub-runner#" + this.id + ": iter[" + this.subIteration + "] | all done.");
      return new Integer(this.subIteration);
    }
  }
  
  private void globalSyncInit() throws SyncException {
    this.runnerLock.lock();
    try {
      if (this.waiters.get() != this.numberThreads) {
        //LOG.info("master init before await " + this.waiters.get());
        this.runnerCondition.await();
        //LOG.info("master init after await " + this.waiters.get());
      }
    } catch (Exception e) {
      throw new SyncException(e.toString());
    } finally {
      this.runnerLock.unlock();
    }
  }
  
  private void globalSync() throws SyncException {
    this.runnerLock.lock();
    try {
      this.waiters.set(0);
      //LOG.info("master before singal all");
      this.subRunnerCondition.signalAll();
      //LOG.info("master singal all");
      
      if (this.waiters.get() != this.numberThreads)
        this.runnerCondition.await();
      //LOG.info("master after await " + this.waiters.get());
    } catch (Exception e) {
      throw new SyncException(e.toString());
    } finally {
      this.runnerLock.unlock();
    }
  }
  
  private void globalSyncFinit() throws SyncException {
    this.runnerLock.lock();
    try {
      this.waiters.set(0);
      this.subRunnerCondition.signalAll();
    } catch (Exception e) {
      throw new SyncException(e.toString());
    } finally {
      this.runnerLock.unlock();
    }
  } 
  
  @SuppressWarnings("unchecked")
  private void initSubRunners(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    List<Vertex<V, E, M>>[] subLists = new List[numberThreads];
    int numberSubVertices = vertices.size() / numberThreads;
    int from = 0, to = 0;
    
    // partition vertices to N subset
    for (int i = 0; i < (numberThreads - 1); i++) {
      from = to;
      to = from + numberSubVertices;
      subLists[i] = vertices.subList(from, to);
      LOG.debug("master: sub-runner#" + i 
            + " sublits:[" + from + "-" + to +"]" 
            + " size=" + subLists[i].size());
    }
    from = to;
    to = vertices.size();
    subLists[numberThreads-1] = vertices.subList(from, to);
    LOG.debug("master: sub-runner#" + (numberThreads-1) 
          + " sublits:[" + from + "-" + to +"]" 
          + " size=" + subLists[(numberThreads-1)].size());
    
    subAliveVertexCounts = new long[numberThreads];
    subActiveVertexCounts = new long[numberThreads];
    subLocalMessageCounts = new long[numberThreads];
    subRemoteMessageCounts = new long[numberThreads];
    
    // create N threads
    threadPool = (ThreadPoolExecutor) Executors.newFixedThreadPool(numberThreads);
    completionService = 
            new ExecutorCompletionService<Integer>(threadPool);
    CyclicBarrier localBarrier = new CyclicBarrier(numberThreads); // local sync

    for (int i = 0; i < numberThreads; i++) {
      completionService.submit(new GraphJobSubRunner(peer, i, 
          subLists[i], localBarrier));
    }
    
    LOG.info("Init " + numberThreads + " sub-runners done.");
  }
  
  private void initRunner(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    bundles = new LyraMessageBundle[peer.getNumPeers()];
    for (int i = 0; i < bundles.length; i++)
      bundles[i] = new LyraMessageBundle();
    peer.registerMessaging(replicas, globalStat);
    LOG.info("Init runner done.");
  }
  
  @Override
  public final void setup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Setup graph!");
    
    setupTOT = System.currentTimeMillis();
    // [SETUP-LD]: load hama graph
    setupFields(peer);
    loadVertices(peer);

    // [SETUP-REP]: build replicated graph
    setupREP = System.currentTimeMillis();
    loadReplicas(peer);
    setupREP = System.currentTimeMillis() - setupREP;
    
    setupREV = System.currentTimeMillis();
    registerReplicas(peer);
    setupREV = System.currentTimeMillis() - setupREV;
    
    // convert vertices Map to List
    this.vertices = new ArrayList<Vertex<V, E, M>>(this.verticesMap.values());
    this.verticesMap = null;
    countGlobal(peer);
    
    // [SETUP-INIT]: initialize all vertices 
    setupINIT = System.currentTimeMillis();
    if (multiThreads) {
      try {
        peer.initilizeLyraMessenger(numberThreads, numberReceivers);
      } catch (ClassNotFoundException e) {
        Assert.fail("ClassNotFoundException  during register replicase!");
      }
      
      initSubRunners(peer);
      globalSyncInit(); // commit, compute and prepare done.
    } else {
      try {
        peer.initilizeLyraMessenger();
      } catch (ClassNotFoundException e) {
        Assert.fail("ClassNotFoundException  during register replicase!");
      }
      
      initRunner(peer);
      commit(peer);
      init(peer);
      peer.prepare();
    }
    
    peer.lyraSync(); // 1st sync
    
    if (multiThreads)
      globalSync(); // send msgs done.
    else 
      send(peer);
    
    System.gc();
    peer.lyraSync(); // 2nd sync
    setupINIT = System.currentTimeMillis() - setupINIT;
    setupTOT = System.currentTimeMillis() - setupTOT;
    setupLD = setupTOT - setupREP - setupREV - setupINIT;    
    
    iteration++;    
  }
  
  @Override
  public final void bsp(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {

    LOG.info("Starting graph processing!");
    
    // until vertices convergence or max iteration
    bspTOT = System.currentTimeMillis();
    while (!isDone()) {
      reset(); // reset global stats
      
      // [BSP-ABT]: collect global stats and arbitrate next step
      onceMST = System.currentTimeMillis();
      arbitrate(peer);
      if (isConverged())  // if converged
        break;
      onceMST = System.currentTimeMillis() - onceMST;
      bspMST += onceMST;

      // [BSP-CMP]: loop over vertices and do commit and computation
      onceSTP = System.currentTimeMillis();
      if (multiThreads) {
        globalSync(); // commit, compute and prepare done.
      } else {
        commit(peer);
        compute(peer);
        peer.prepare();
      }
      onceSTP = System.currentTimeMillis() - onceSTP;
      bspSTP += onceSTP;
      
      // [BSP-SYN1]: sync with all peers
      onceSYN = System.currentTimeMillis();
      peer.lyraSync();  // 1st sync
      onceSYN = System.currentTimeMillis() - onceSYN;
      bspSYN += onceSYN;
      
      // [BSP-SND]: send sync and stat msgs
      onceSND = System.currentTimeMillis();
      if (multiThreads) {
        globalSync(); 
        // transfer done.
      } else {
        send(peer);
      }
      onceSND = System.currentTimeMillis() - onceSND;
      bspSND += onceSND;
      
      // [BSP-SYN2]: sync with all workers
      onceSYN2 = System.currentTimeMillis();
      peer.lyraSync();  // 2nd sync
      onceSYN2 = System.currentTimeMillis() - onceSYN2;
      bspSYN += onceSYN2;
      
      onceTOT = onceMST + onceSTP + onceSYN + onceSND + onceSYN2;
      
      // [BSP-DMP]: dump superstep info
      if (isMasterTask(peer)) {
        LOG.info("==== iter#" + iteration + ":"
            + " [msg] #rcv-sync: " + globalSyncMessageCount
            + " #snd-local: " + globalLocalMessageCount
            + " #snd-remote: " + globalRemoteMessageCount
            + " [vtx] #active: " + globalActiveVertexCount
            + " #alive: " + globalAliveVertexCount
            + " [cvg] rate(%): " + (this.rate * 100f) + " ====");
      }
      
      LOG.info("==== iter#" + iteration + ":" +
            " [tot] " + onceTOT + 
            " [mst] " + onceMST + 
            " [cmp] " + onceSTP + 
            " [snd] " + onceSND + 
            " [syn] " + (onceSYN + onceSYN2) + " ("+ onceSYN + " + " + onceSYN2 + ")" +
            " ==== \n");
      
      this.iteration++;
      if (isMasterTask(peer)) {
        peer.getCounter(GraphJobCounter.ITERATIONS).increment(1);
      }
    }
    //bspTOT = System.currentTimeMillis() - bspTOT;
    bspTOT = bspSYN + bspPRS + bspMST + bspSTP + bspSND;
    
    if (multiThreads)
      globalSyncFinit();
    
    if (isDone())
      LOG.info("BSP Job Done: max-iter (" + iteration + ") rate(%): " + (this.rate * 100f));   
    else
      LOG.info("BSP Job Done: converged (" + iteration + ") rate(%): " + (this.rate * 100f));

    average(peer);
    breakdown();
    //rpctest(peer);
    jtest(peer);
    //synctest(peer);
    //msgtest(peer);
    //msgtest2(peer);
  }
  
  /**
   * Just write <ID as Writable, Value as Writable> pair as a result. Note that
   * this will also be executed when failure happened.
   */
  @Override
  public final void cleanup(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    if (multiThreads) {
      for (int i = 0; i < this.numberThreads; i++) {
        try {
          Future<Integer> take = completionService.take();
          if (take != null) {
            LOG.info("sub-runner#" + i + ": finish " + take.get() + " iterations");
          }
        } catch (Exception e) {
          LOG.error("Exception during cleanup sub runners!", e);
          break;
        }
      }
      threadPool.shutdownNow();
    }
    
    for (Vertex<V, E, M> e : vertices) {
      peer.write(e.getVertexID(), e.getNewValue()); // (BSP-g)
    }
  }

  private final void breakdown() {
    LOG.info("Peer#" + peer.getPeerIndex());
    LOG.info("==== setup tot: " + setupTOT + 
          " load: " + setupLD + 
          " rep: " + setupREP +
          " rev: " + setupREV +
          " init: " + setupINIT + " ====");
    LOG.info("====   bsp tot: " + bspTOT + 
          " syn: " + bspSYN + 
          " prs: " + bspPRS + 
          " mst: " + bspMST + 
          " stp: " + bspSTP + 
          " snd: " + bspSND + " ====");
    LOG.info("====   msg tot: " + msgRCV + " ====");
    
    if (isMasterTask(peer)) {
      LOG.info("Average");
      LOG.info("==== setup tot: " + avgSetupTOT + 
          " load: " + avgSetupLD + 
          " rep: " + avgSetupREP +
          " rev: " + avgSetupREV +
          " init: " + avgSetupINIT + " ====");
      LOG.info("====   bsp tot: " + avgBspTOT + 
          " syn: " + avgBspSYN + 
          " prs: " + avgBspPRS + 
          " mst: " + avgBspMST + 
          " stp: " + avgBspSTP + 
          " snd: " + avgBspSND + " ====");
      
      LOG.info("Sum");
      LOG.info("====   msg tot: " + sumMsgTOT + " ====");
    }
  }
  
  
  private final void average(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    int peerIndex = peer.getPeerIndex();
    if (multiThreads)
      peer.prepare(0);
    else
      peer.prepare();
    peer.lyraSync();
    
    localStat.put(new Text("bsp-tot" + peerIndex), new LongWritable(bspTOT));
    localStat.put(new Text("bsp-syn" + peerIndex), new LongWritable(bspSYN));
    localStat.put(new Text("bsp-prs" + peerIndex), new LongWritable(bspPRS));
    localStat.put(new Text("bsp-mst" + peerIndex), new LongWritable(bspMST));
    localStat.put(new Text("bsp-stp" + peerIndex), new LongWritable(bspSTP));
    localStat.put(new Text("bsp-snd" + peerIndex), new LongWritable(bspSND));
    
    localStat.put(new Text("setup-tot" + peerIndex), new LongWritable(setupTOT));
    localStat.put(new Text("setup-ld" + peerIndex), new LongWritable(setupLD));
    localStat.put(new Text("setup-rep" + peerIndex), new LongWritable(setupREP));
    localStat.put(new Text("setup-rev" + peerIndex), new LongWritable(setupREV));
    localStat.put(new Text("setup-init" + peerIndex), new LongWritable(setupINIT));
    
    if (multiThreads)
      peer.transfer(0, getMasterTask(peer), localStat);
    else
      peer.transfer(getMasterTask(peer), localStat);
    localStat.clear();
    
    peer.lyraSync();
    
    if (isMasterTask(peer)) {
      long btot = 0, bsyn = 0, bprs = 0, bmst = 0, bstp = 0, bsnd = 0;
      long stot = 0, sld = 0, srep = 0, srev = 0, sinit = 0;
      int numberPeers = peer.getNumPeers();

      for (Entry<Writable, Writable> e : globalStat.entrySet()) {
        String key = e.getKey().toString();

        if (key.startsWith("bsp-tot")) {
          btot += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-syn")) {
          bsyn += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-prs")) {
          bprs += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-mst")) {
          bmst += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-stp")) {
          bstp += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("bsp-snd")) {
          bsnd += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-tot")) {
          stot += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-ld")) {
          sld += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-rep")) {
          srep += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-rev")) {
          srev += ((LongWritable) e.getValue()).get();
        } else if (key.startsWith("setup-init")) {
          sinit += ((LongWritable) e.getValue()).get();
        } else{
          throw new UnsupportedOperationException("unknown time: " + key);
        }
      }
      
      avgBspTOT = btot / numberPeers;
      avgBspSYN = bsyn / numberPeers;
      avgBspPRS = bprs / numberPeers;
      avgBspMST = bmst / numberPeers;
      avgBspSTP = bstp / numberPeers;
      avgBspSND = bsnd / numberPeers;
      
      avgSetupTOT = stot / numberPeers;
      avgSetupLD = sld / numberPeers;
      avgSetupREP = srep / numberPeers;
      avgSetupREV = srev / numberPeers;
      avgSetupINIT = sinit / numberPeers;
    }
  }
  
  private final void msgtest(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    Assert.assertFalse(multiThreads);
    LyraTestMessageBundle xbundle = new LyraTestMessageBundle();
    LyraTestMessage msg = null;
    int count = 5000000;
    int offset = peer.getPeerIndex() * count;
    int n = 2500, piece = count/n;
    double[] testSet = new double[count*peer.getNumPeers()];
    long time=0;
    
    LOG.info("Init message testing.");
    
    if (isMasterTask(peer)) {
      peer.registerMessageTesting(testSet);
    }
    
    peer.prepare();
    peer.lyraSync();
    time = System.currentTimeMillis();
    
    if (!isMasterTask(peer)) {
      for (int j = 0; j < n-1; j++) {
        int add = j*piece;
        for (int i = 0; i < piece; i++) {
          msg = new LyraTestMessage(new LongWritable(i + add + offset), new DoubleWritable((1.0)));
          xbundle.addMessage(msg);
        }
        peer.transfer(getMasterTask(peer), xbundle);
        xbundle.clear();
      }
      
      for (int i = piece*(n-1); i < count; i++) {
        msg = new LyraTestMessage(new LongWritable(i + offset), new DoubleWritable((1.0)));
        xbundle.addMessage(msg);
      }
      peer.transfer(getMasterTask(peer), xbundle);
      
    }
    
    peer.lyraSync();
    time = System.currentTimeMillis() - time;
    if (isMasterTask(peer)) {
      LOG.info("Message Testing: " + time + " usecs serve " + peer.getNumCurrentLyraMessages() + " messages");
    }
  }
  
  private final void msgtest2(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    Assert.assertFalse(multiThreads);
    int count = 5000000, recv = 0;
    int offset = peer.getPeerIndex() * count;
    double[] testSet = new double[count*peer.getNumPeers()];
    long time=0, snd=0, prs=0;
        
    LOG.info("Init message testing.");
    
    peer.sync();
    time = System.currentTimeMillis();
    
    if (!isMasterTask(peer)) {
      for (int i = 0; i < count; i++) {
        peer.send(getMasterTask(peer), new GraphJobMessage(
            new LongWritable(i+offset), new DoubleWritable(1.0)));
      }
    }
    
    peer.sync();
    snd = System.currentTimeMillis() - time;
    time = System.currentTimeMillis();
    
    if (isMasterTask(peer)) {
      LOG.info("get" + peer.getNumCurrentMessages() + " messages.");
      
      GraphJobMessage msg = null;
      while ((msg = peer.getCurrentMessage()) != null) {
        testSet[(int)msg.getID().get()] = msg.getValue().get();
        recv++;
      }
      
      prs = System.currentTimeMillis() - time;
      LOG.info("Message Testing (" + recv + "): snd = " + snd + " prs = " + prs);
    }
  }
  
  private final void synctest(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    //DoubleWritable value = new DoubleWritable(1.0d);
    Assert.assertFalse(multiThreads);
    
    int count = 100;
    
    peer.prepare();
    peer.lyraSync();
    
    long time = System.currentTimeMillis();
    for (int i = 0; i < count; i++) {
      peer.prepare();
      peer.lyraSync();
    }
    time = System.currentTimeMillis() - time;
    
    if (isMasterTask(peer)) {
      LOG.info("Spend " + time + " usecs to sync " + count + " times ");
    }
  }
  
  private final void rpctest(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    //DoubleWritable value = new DoubleWritable(1.0d);
    Assert.assertFalse(multiThreads);
    int count = 100000;
    MapWritable map = new MapWritable();
    map.put(new Text("hello"), new LongWritable(100));
    
    peer.prepare();
    peer.lyraSync();
    long time = System.currentTimeMillis();
    
    if (!isMasterTask(peer)) {
      for (int i = 0; i < count; i++) {
        //peer.transfer(getMasterTask(peer));
        peer.transfer(getMasterTask(peer), map);
      }
    }
    
    peer.lyraSync();
    time = System.currentTimeMillis() - time;
    
    if (isMasterTask(peer)) {
      long number = peer.getNumCurrentLyraMessages();
      LOG.info("RPC test: spend " + time/1000 + " secs to invoke " + number + " RPC call ");
    }
  }
  
  private final void jtest(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException, SyncException, InterruptedException{
    double[] neighbours = new double[100];
    double lastValue = 0.0d;
    int count = 100000000;
    for (int i = 0; i < 100; i++)
      neighbours[i] = 1.0d /(double)i;
    
    long time = System.currentTimeMillis();
    
    for (int i = 0 ; i < count; i++) {
      double sum = 0;
      for (int j = 0; j < 100; j++) {
        sum += neighbours[j];
      }
      double value = (1.0d - 0.85d) / 100 + (0.85 * sum);
      
      lastValue += value;
    }
    
    time = System.currentTimeMillis() - time;
    LOG.info("Use " +time + " secs Last value = " + lastValue + " xx=" + count);
  }
  
  private final boolean isDone() {
    return (this.maxIteration > 0) && (this.iteration > this.maxIteration);
  }
  
  private final boolean isConverged() {
    return (this.converged ||
        (this.minRate > 0.0f) && (this.rate < this.minRate));
  }

  private final void reset() {
    // recieve
    globalActiveVertexCount = 0;
    globalAliveVertexCount = 0;
    globalLocalMessageCount = 0;
    globalRemoteMessageCount = 0;
    globalSyncMessageCount = 0;
    this.converged = true;
    
    // send
    localActiveVertexCount = 0;
    localAliveVertexCount = 0;
    localLocalMessageCount = 0;
    localRemoteMessageCount = 0;
    localSyncMessageCount = 0;    
  }
  
  private void arbitrate(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    if (multiThreads)
      for (int i = 0; i < numberThreads; i++)
        localSyncMessageCount += peer.getNumCurrentLyraMessages(i);
    else
      localSyncMessageCount = peer.getNumCurrentLyraMessages();
    localSyncMessageCount -= peer.getNumPeers(); // minus the global-stat msgs from each peer
    
    for (Entry<Writable, Writable> e : globalStat.entrySet()) {
      String key = e.getKey().toString();
      if (key.startsWith(S_FLAG_HALT)) {
        if (((BooleanWritable) e.getValue()).get() == false)
          this.converged = false;
      } else if (key.startsWith(S_FLAG_ALIVE_COUNTS)) {
        globalAliveVertexCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_LOCAL_MESSAGE_COUNTS)) {
        globalLocalMessageCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_REMOTE_MESSAGE_COUNTS)) {
        globalRemoteMessageCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_LAST_ACTIVE_COUNTS)) {
        globalActiveVertexCount += ((LongWritable) e.getValue()).get();
      } else if (key.startsWith(S_FLAG_SYNC_MESSAGE_COUNTS)) {
        globalSyncMessageCount += ((LongWritable) e.getValue()).get();
      } else {
        throw new UnsupportedOperationException("unknown stat: " + key);
      }
    }
    
    if (aggregationRunner.isEnabled())
      // update aggregation value if necessary (additional sync)
      throw new UnsupportedOperationException("aggregation unsupported now!");

    // TODO: it is equal to the measure schema to globalError used by Hama
    this.rate = ((float)this.globalAliveVertexCount / (float)this.numberVertices);
    //this.rate = ((float)this.globalActiveVertexCount / (float)this.numberVertices);
    
    sumMsgTOT += globalSyncMessageCount;
    msgRCV += localSyncMessageCount;

    LOG.debug("arbitrate #sync-msgs: " + localSyncMessageCount 
          + " rate(%): " + (this.rate*100f));
  }
  
  private void commit(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    for (Vertex<V, E, M> vertex : vertices)
      vertex.commit();
  }
  /**
   * Do the main logic of a superstep, namely checking if vertices are active,
   * feeding compute with messages and controlling combiners/aggregators.
   */
  private void compute(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException {
    long activeVertexCount = 0;
    long aliveVertexCount = 0;
    long localMessageCount = 0;
    long remoteMessageCount = 0;
    
    for (Vertex<V, E, M> vertex : vertices) {
      if (!vertex.isHalted()) {
        activeVertexCount++;
        
        //M lastValue = vertex.getLastValue();
        //TODO: modify compute function interface?
        vertex.compute(null);
        localMessageCount += vertex.localSync();
        remoteMessageCount += vertex.remoteSync(bundles);

        //aggregationRunner.aggregateVertex(lastValue, vertex);

        if (!vertex.isConverged())
          aliveVertexCount++;
      }
    }
    
    // [unused] send aggregation value if necessary (need additional sync)
    //aggregationRunner.sendAggregatorValues(peer);
    
    localActiveVertexCount = activeVertexCount;
    localAliveVertexCount = aliveVertexCount;
    localLocalMessageCount = localMessageCount;
    localRemoteMessageCount = remoteMessageCount;
  }

  /**
   * Seed the vertices first with their own values in compute. This is the first
   * superstep after the vertices have been loaded.
   */
  private void init(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException {
    long activeVertexCount = 0;
    long aliveVertexCount = 0;
    long localMessageCount = 0;
    long remoteMessageCount = 0;
    
    for (Vertex<V, E, M> vertex : vertices) {
      activeVertexCount++;
      
      //M lastValue = vertex.getLastValue(); // (BSP-g)
      //TODO: modify compute function interface?
      vertex.compute(null);
      localMessageCount += vertex.localSync();
      remoteMessageCount += vertex.remoteSync(bundles);

      //aggregationRunner.aggregateVertex(lastValue, vertex);
      
      if (!vertex.isConverged())
        aliveVertexCount++;
    }

    // [unused] send aggregation value if necessary (need additional sync)
    //aggregationRunner.sendAggregatorValues(peer);
    
    localActiveVertexCount = activeVertexCount;
    localAliveVertexCount = aliveVertexCount;
    localLocalMessageCount = localMessageCount;
    localRemoteMessageCount = remoteMessageCount;
  }
  
  private void send(BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    /*for (int i = 0; i < bundles.length; i++) {
      if (!bundles[i].isEmpty()) {
        peer.transfer(peer.getPeerName(i), bundles[i]);
        bundles[i] = new LyraMessageBundle();
      }
    }*/
    
    for (int i = 0; i < bundles.length; i++) {
      if (!bundles[i].isEmpty()) {
        int n = bundles[i].size() / bundleSize;
        if (n > 1) {
          List<LyraMessageBundle> list = bundles[i].split(n);
          
          for (LyraMessageBundle bundle: list)
            peer.transfer(peer.getPeerName(i), bundle);
        } else {
          peer.transfer(peer.getPeerName(i), bundles[i]);
        }
        
        bundles[i] = new LyraMessageBundle();
      }
    }
    
    // send local-stat to all peers
    if ((localAliveVertexCount>0) || (localLocalMessageCount>0) || (localRemoteMessageCount>0))
      localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(false));
    else 
      localStat.put(GraphJobRunner.FLAG_STAT_HALT, new BooleanWritable(true));
    localStat.put(GraphJobRunner.FLAG_STAT_LAST_ACTIVE_COUNTS, new LongWritable(localActiveVertexCount));
    localStat.put(GraphJobRunner.FLAG_STAT_ALIVE_COUNTS, new LongWritable(localAliveVertexCount));
    localStat.put(GraphJobRunner.FLAG_STAT_LOCAL_MESSAGE_COUNTS, new LongWritable(localLocalMessageCount));
    localStat.put(GraphJobRunner.FLAG_STAT_REMOTE_MESSAGE_COUNTS, new LongWritable(localRemoteMessageCount));
    localStat.put(GraphJobRunner.FLAG_STAT_SYNC_MESSAGE_COUNTS, new LongWritable(localSyncMessageCount));
    
    for (String peerName : peer.getAllPeerNames())
      peer.transfer(peerName, localStat);
    localStat.clear();

    LOG.debug("worker:"
        + " [msg] #local = " + localLocalMessageCount
        + " #remote = " + localRemoteMessageCount
        + " [vtx] #active: " + localActiveVertexCount
        + " #alive: " + localAliveVertexCount
        + " [rep] #synced: " + localSyncMessageCount);
  }

  @SuppressWarnings("unchecked")
  private void setupFields(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    this.peer = peer;
    this.conf = peer.getConfiguration();
    
    maxIteration = peer.getConfiguration().getInt("hama.graph.max.iteration",-1);
    minRate = peer.getConfiguration().getFloat("hama.graph.min.rate", -1f);
    numberThreads = peer.getConfiguration().getInt("hama.graph.mun.threads", 1);
    numberReceivers = peer.getConfiguration().getInt("hama.graph.mun.receivers", numberThreads);
    if (numberThreads > 1)
      this.multiThreads = true;
    LOG.info("Start " + numberThreads + " threads and " + numberReceivers + " receivers" );
    
    vertexIdClass = (Class<V>) conf.getClass(GraphJob.VERTEX_ID_CLASS_ATTR,
        Text.class, Writable.class);
    vertexValueClass = (Class<M>) conf.getClass(
        GraphJob.VERTEX_VALUE_CLASS_ATTR, IntWritable.class, Writable.class);
    edgeValueClass = (Class<E>) conf.getClass(
        GraphJob.VERTEX_EDGE_VALUE_CLASS_ATTR, IntWritable.class,
        Writable.class);
    vertexClass = (Class<Vertex<V, E, M>>) conf.getClass(
        "hama.graph.vertex.class", Vertex.class);

    // set the classes statically, so we can save memory per message
    GraphJobMessage.VERTEX_ID_CLASS = vertexIdClass;
    GraphJobMessage.VERTEX_VALUE_CLASS = vertexValueClass;
    GraphJobMessage.VERTEX_CLASS = vertexClass;
    GraphJobMessage.EDGE_VALUE_CLASS = edgeValueClass;
    
    LyraSyncMessage.VERTEX_ID_CLASS = vertexIdClass;
    LyraSyncMessage.VERTEX_VALUE_CLASS = vertexValueClass;

    partitioner = (Partitioner<V, M>) ReflectionUtils.newInstance(
        conf.getClass("bsp.input.partitioner.class", HashPartitioner.class),
        conf);
    partitioner.setConf(conf);

    if (!conf.getClass(MESSAGE_COMBINER_CLASS, Combiner.class).equals(
        Combiner.class)) {
      LOG.debug("Combiner class: " + conf.get(MESSAGE_COMBINER_CLASS));

      combiner = (Combiner<M>) ReflectionUtils.newInstance(
          conf.getClass(MESSAGE_COMBINER_CLASS, Combiner.class),
          conf);
    }

    aggregationRunner = new AggregationRunner<V, E, M>();
    aggregationRunner.setupAggregators(peer);
    
    // unify stat name
    FLAG_STAT_HALT = new Text(S_FLAG_HALT 
        + "-" + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_ALIVE_COUNTS = new Text(S_FLAG_ALIVE_COUNTS 
        + "-" + Integer.toString(peer.getPeerIndex()));   
    FLAG_STAT_LOCAL_MESSAGE_COUNTS = new Text(S_FLAG_LOCAL_MESSAGE_COUNTS 
        + "-" + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_REMOTE_MESSAGE_COUNTS = new Text(S_FLAG_REMOTE_MESSAGE_COUNTS 
        + "-" + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_LAST_ACTIVE_COUNTS = new Text(S_FLAG_LAST_ACTIVE_COUNTS 
        + "-" + Integer.toString(peer.getPeerIndex()));
    FLAG_STAT_SYNC_MESSAGE_COUNTS = new Text(S_FLAG_SYNC_MESSAGE_COUNTS 
        + "-" + Integer.toString(peer.getPeerIndex()));
    
    // TODO: Now, we only provide partitioning code
    runtimePartitioning = conf.getBoolean(
        GraphJob.VERTEX_GRAPH_RUNTIME_PARTIONING, true);
    Assert.assertTrue(runtimePartitioning);
    
    selectiveActivate = conf.getBoolean(
        GRAPH_SELECTIVE_ACTIVATE, true);
    LOG.info("Selective activation is enabled");
  }

  /**
   * Partitions our vertices through multiple supersteps to save memory.
   */
  private int partitionMultiSteps(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      long splitSize) throws IOException, SyncException, InterruptedException {
    int multiSteps = 1;

    MapWritable ssize = new MapWritable();
    ssize
        .put(new IntWritable(peer.getPeerIndex()), new LongWritable(splitSize));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      long maxSplitSize = 0L;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          long curr = ((LongWritable) e.getValue()).get();
          if (maxSplitSize < curr) {
            maxSplitSize = curr;
          }
        }
      }

      int steps = (int) (maxSplitSize / conf.getLong( // 20 mb
          "hama.graph.multi.step.partitioning.interval", 20000000)) + 1;

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("max"), new IntWritable(steps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    if (isMasterTask(peer)) {
      peer.getCounter(GraphJobCounter.MULTISTEP_PARTITIONING).increment(
          multiSteps);
    }

    return multiSteps;
  }

  @SuppressWarnings("unchecked")
  private void createVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) 
      throws IOException {
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      Vertex<V, E, M> vertex = (Vertex<V, E, M>) msg.getVertex();
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);
      
      Vertex<V, E, M> old = verticesMap.put(vertex.getVertexID(), vertex);
      if (old != null) // vertex with partitioned info
        vertex.merge(old);
      
      vertex.setup(conf); // call setup after merge
      
      if (selectiveActivate)
        vertex.enableSelectiveActivate();
    }
  }
  
  /**
   * Loads vertices into memory of each peer. TODO this needs to be simplified.
   */
  @SuppressWarnings("unchecked")
  private void loadVertices(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start loading vertices");
    
    final VertexInputReader<Writable, Writable, V, E, M> reader = 
        (VertexInputReader<Writable, Writable, V, E, M>) ReflectionUtils
        .newInstance(conf.getClass(GraphJob.VERTEX_GRAPH_INPUT_READER,
            VertexInputReader.class), conf);

    final boolean repairNeeded = conf.getBoolean(GRAPH_REPAIR, false);
    final long splitSize = peer.getSplitSize();
    final int partitioningSteps = partitionMultiSteps(peer, splitSize);
    multiSteps = partitioningSteps; // used by load replicas
    final long interval = splitSize / partitioningSteps;
    final boolean selfReference = conf.getBoolean("hama.graph.self.ref", false);

    // new vertex container
    Vertex<V, E, M> vertex = newVertexInstance(vertexClass, conf);
    vertex.runner = this;
    vertex.setVertexType(VertexType.MASTER);

    long startPos = peer.getPos();
    if (startPos == 0)
      startPos = 1L;

    KeyValuePair<Writable, Writable> next = null;
    int numberPeers = peer.getNumPeers();
    boolean pending = false; // fix the tail-vertex issue
    int steps = 1;
    
    while ((next = peer.readNext()) != null) {
      boolean finished = false;
      try {
        finished = reader.parseVertex(next.getKey(), next.getValue(), vertex);
        pending = true;
      } catch (Exception e) {
        throw new IOException("exception occured during parsing vertex!"
            + e.toString());
      }

      if (!finished)
        continue;

      if (vertex.getEdges() == null) {
        if (selfReference) {
          vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
              .getVertexID(), null)));
        } else {
          vertex.setEdges(Collections.EMPTY_LIST);
        }
      }

      if (selfReference)
        vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));

      if (runtimePartitioning) {
        int partition = partitioner.getPartition(vertex.getVertexID(),
            vertex.getNewValue(), numberPeers);
        peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
      } else {
        vertex.setup(conf);
        verticesMap.put(vertex.getVertexID(), vertex);
      }
      
      // new vertex container
      vertex = newVertexInstance(vertexClass, conf);
      vertex.runner = this;
      vertex.setVertexType(VertexType.MASTER);

      if (runtimePartitioning) {
        if (steps < partitioningSteps && (peer.getPos() - startPos) >= interval) {
          peer.sync();
          steps++;
          createVertices(peer);
          startPos = peer.getPos();
        }
      }
    }

    // flush the tail-vertex
    if (pending) {
      boolean tail = false;
      try {
        tail = reader.parseVertex(null, null, vertex);
      } catch (Exception e) {
        throw new IOException("exception occured during parsing tail vertex!"
            + e.toString());
      }
      
      if (tail) {
        if (vertex.getEdges() == null) {
          if (selfReference) {
            vertex.setEdges(Collections.singletonList(new Edge<V, E>(vertex
                .getVertexID(), null)));
          } else {
            vertex.setEdges(Collections.EMPTY_LIST);
          }
        }
  
        if (selfReference)
          vertex.addEdge(new Edge<V, E>(vertex.getVertexID(), null));
        
        if (runtimePartitioning) {
          int partition = partitioner.getPartition(vertex.getVertexID(),
              vertex.getNewValue(), peer.getNumPeers());
          peer.send(peer.getPeerName(partition), new GraphJobMessage(vertex));
        } else {
          vertex.setup(conf);
          verticesMap.put(vertex.getVertexID(), vertex);
        }
        
        LOG.debug("Oops! Fix the tail-vertex(" + vertex.getVertexID() + ") issue.");
      }
    }

    // last step
    if (runtimePartitioning) {
      peer.sync();
      createVertices(peer);
    }

    LOG.info("Load vertices finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());

    /*
     * If the user want to repair the graph, it should traverse through that
     * local chunk of adjancency list and message the corresponding peer to
     * check whether that vertex exists. In real-life this may be dead-ending
     * vertices, since we have no information about outgoing edges. Mainly this
     * procedure is to prevent NullPointerExceptions from happening.
     */
    if (repairNeeded)
      repair(peer, partitioningSteps, selfReference);
  }
  
  @SuppressWarnings("unchecked")
  private void repair(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer,
      int partitioningSteps, boolean selfReference) throws IOException,
      SyncException, InterruptedException {
    LOG.info("Start repairing the graph!");
    
    int multiSteps = 0;
    MapWritable ssize = new MapWritable();
    ssize.put(new IntWritable(peer.getPeerIndex()),
        new IntWritable(verticesMap.size()));
    peer.send(getMasterTask(peer), new GraphJobMessage(ssize));
    ssize = null;
    peer.sync();

    if (isMasterTask(peer)) {
      int minVerticesSize = Integer.MAX_VALUE;
      GraphJobMessage received = null;
      while ((received = peer.getCurrentMessage()) != null) {
        MapWritable x = received.getMap();
        for (Entry<Writable, Writable> e : x.entrySet()) {
          int curr = ((IntWritable) e.getValue()).get();
          if (minVerticesSize > curr) {
            minVerticesSize = curr;
          }
        }
      }
      
      if (minVerticesSize < (partitioningSteps * 2)) {
        multiSteps = minVerticesSize;
      } else {
        multiSteps = (partitioningSteps * 2);
      }

      for (String peerName : peer.getAllPeerNames()) {
        MapWritable temp = new MapWritable();
        temp.put(new Text("steps"), new IntWritable(multiSteps));
        peer.send(peerName, new GraphJobMessage(temp));
      }
    }
    peer.sync();

    GraphJobMessage received = peer.getCurrentMessage();
    MapWritable x = received.getMap();
    for (Entry<Writable, Writable> e : x.entrySet()) {
      multiSteps = ((IntWritable) e.getValue()).get();
    }

    Map<V, Vertex<V, E, M>> tmp = new HashMap<V, Vertex<V, E, M>>();

    int i = 0;
    int syncs = 0;

    for (Vertex<V, E, M> v : verticesMap.values()) {
      for (Edge<V, E> e : v.getEdges()) {
        peer.send(v.getDestinationPeerName(e),
            new GraphJobMessage(e.getDestinationVertexID()));
      }

      if (syncs < multiSteps && (i % (verticesMap.size() / multiSteps)) == 0) {
        peer.sync();
        syncs++;
        GraphJobMessage msg = null;
        while ((msg = peer.getCurrentMessage()) != null) {
          V vertexName = (V) msg.getVertexId();

          Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
          newVertex.setVertexID(vertexName);
          newVertex.runner = this;
          if (selfReference) {
            newVertex.setEdges(Collections.singletonList(new Edge<V, E>(
                newVertex.getVertexID(), null)));
          } else {
            newVertex.setEdges(new ArrayList<Edge<V, E>>(0));
          }
          newVertex.setup(conf);
          tmp.put(vertexName, newVertex);

        }
      }
      i++;
    }

    peer.sync();
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      V vertexName = (V) msg.getVertexId();

      Vertex<V, E, M> newVertex = newVertexInstance(vertexClass, conf);
      newVertex.setVertexID(vertexName);
      newVertex.runner = this;
      if (selfReference) {
        newVertex.setEdges(Collections.singletonList(new Edge<V, E>(newVertex
            .getVertexID(), null)));
      } else {
        newVertex.setEdges(new ArrayList<Edge<V, E>>(0));
      }
      newVertex.setup(conf);
      tmp.put(vertexName, newVertex);
      newVertex = null;

    }

    for (Vertex<V, E, M> e : verticesMap.values()) {
      if (tmp.containsKey((e.getVertexID()))) {
        tmp.remove(e.getVertexID());
      }
    }

    verticesMap.putAll(tmp);
    tmp.clear();
    
    LOG.info("Repair finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local vertices: " + verticesMap.size());
  }

  @SuppressWarnings("unchecked")
  private void createReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isReplicaMessage(), true);
      
      V vertexId = (V) msg.getDestinationVertexId();
      V sourceVertexId = (V) msg.getVertexId();
      M vertexValue = (M) msg.getVertexValue();
      E edgeValue = (E) msg.getEdgeValue();
      int peerIndex = msg.getPeerIndex().get();
      
      Vertex<V, E, M> replica;
      if (!replicasMap.containsKey(sourceVertexId)) {
        // create replica
        replica = newVertexInstance(vertexClass, conf);
        replica.runner = this;
        replica.setVertexType(VertexType.REPLICA);
        replica.setVertexID(sourceVertexId);
        replica.setLastValue(vertexValue);
        replica.setMasterPeerIndex(peerIndex);
        // replica does not need to call setup() 
        
        replicasMap.put(sourceVertexId, replica);
      } else {
        replica = replicasMap.get(sourceVertexId);
      }
      
      Vertex<V, E, M> destinationVertex = verticesMap.get(vertexId);
      RefEdge<V, E, M> in = new RefEdge<V, E, M>(replica, edgeValue, RefEdgeType.INEDGE);
      destinationVertex.addInEdge(in);
      RefEdge<V, E, M> out = new RefEdge<V, E, M>(destinationVertex, edgeValue, RefEdgeType.OUTEDGE);
      replica.addOutEdge(out);
      in.setPair(out);
      out.setPair(in);
    }
  }
  
  // load replicas (fill in-edges and local out-edges)
  // TODO: selfReference issue
  @SuppressWarnings("unchecked")
  private void loadReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start loading replicas");
    
    int sourcePeerIndex = peer.getPeerIndex();
    int step = 1, replicaSteps = multiSteps;
    int count = 0, interval = verticesMap.size() / replicaSteps;

    // send msgs to destination vertices
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      for (Edge<V, E> edge : vertex.getEdges()) {
        int destinationPeerIndex = partitioner.getPartition(
            edge.getDestinationVertexID(), (M) edge.getValue(),
            peer.getNumPeers());
        if (selectiveActivate)
          vertex.addNeighborMap(edge.getDestinationVertexID(), destinationPeerIndex);
        
        if (sourcePeerIndex == destinationPeerIndex) { // in local peer
          Vertex<V, E, M> destinationVertex = verticesMap.get(edge.getDestinationVertexID());
          
          RefEdge<V, E, M> in = new RefEdge<V, E, M>(vertex, edge.getValue(), RefEdgeType.INEDGE);
          destinationVertex.addInEdge(in);
          RefEdge<V, E, M> out = new RefEdge<V, E, M>(destinationVertex, edge.getValue(), RefEdgeType.OUTEDGE);
          vertex.addOutEdge(out);
          in.setPair(out);
          out.setPair(in);
        } else { // in remote peer
          String peerName = peer.getPeerName(destinationPeerIndex);
          peer.send(peerName, new GraphJobMessage(vertex.getVertexID(), vertex.getNewValue(),
              edge.getDestinationVertexID(), edge.getValue(), new IntWritable(peer.getPeerIndex())));
        }
      }
      
      count++;
      if (step < replicaSteps && count >= interval) {
        peer.sync();
        createReplicas(peer);
        step++;
        count = 0;
      }
    }
    Assert.assertEquals(step, replicaSteps);
    
    //last step
    peer.sync();
    createReplicas(peer);
    
    LOG.info("Load replicas finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local replicas: " + replicasMap.size());
  }

  // [unused] single step
  private void loadReplicas2(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    int sourcePeerIndex = peer.getPeerIndex();
    
    // send msgs to destination vertices
    for (Vertex<V, E, M> vertex : verticesMap.values()) {
      for (Edge<V, E> edge : vertex.getEdges()) {
        int destinationPeerIndex = partitioner.getPartition(
            edge.getDestinationVertexID(), (M) edge.getValue(),
            peer.getNumPeers());
        
        if (sourcePeerIndex == destinationPeerIndex) { // local peer
          Vertex<V, E, M> destinationVertex = verticesMap.get(edge.getDestinationVertexID());
          
          RefEdge<V, E, M> in = new RefEdge<V, E, M>(vertex, edge.getValue(), RefEdgeType.INEDGE);
          destinationVertex.addInEdge(in);
          RefEdge<V, E, M> out = new RefEdge<V, E, M>(destinationVertex, edge.getValue(), RefEdgeType.OUTEDGE);
          vertex.addOutEdge(out);
          in.setPair(out);
          out.setPair(in);
        } else { // remote peer
          String peerName = peer.getPeerName(destinationPeerIndex);
          peer.send(peerName, new GraphJobMessage(vertex.getVertexID(), vertex.getNewValue(),
              edge.getDestinationVertexID(), edge.getValue(), new IntWritable(peer.getPeerIndex())));
        }
      }
    }

    // create replicas 
    peer.sync();
    createReplicas(peer);
    
    LOG.info("Load replicas finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of local replicas: " + replicasMap.size());
  }
  
  @SuppressWarnings("unchecked")
  private void registerReplicas(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    LOG.info("Start registering replicas");
    int registerCount = 0;
    int step = 1, registerSteps = (multiSteps-1)/2 + 1;
    int count = 0, interval = replicasMap.size() / registerSteps;

    // transfer replicasMap to replicas[], and send register message to master
    this.replicas = (Vertex<V, E, M>[]) Array.newInstance(
        GraphJobMessage.VERTEX_CLASS, this.replicasMap.size());
    long localID = 0;
    for (Vertex<V, E, M> vertex: replicasMap.values()) {
      vertex.setLocalID(localID);
      peer.send(peer.getPeerName(vertex.getMasterPeerIndex()), 
          new GraphJobMessage(new IntWritable(peer.getPeerIndex()), 
          vertex.getVertexID(), new LongWritable((long)localID)));
      
      replicas[(int)localID++] = vertex;
      
      count++;
      if (step < registerSteps && count >= interval) {
        peer.sync();
        GraphJobMessage msg = null;
        long remoteID = 0;
        registerCount += peer.getNumCurrentMessages();
        while ((msg = peer.getCurrentMessage()) != null) {
          Assert.assertEquals(msg.isRegisterMessage(), true);
          
          int peerIndex = msg.getPeerIndex().get();
          V vertexID = (V) msg.getVertexId();
          remoteID = msg.getLocalID().get();
          
          Vertex<V, E, M> master = verticesMap.get(vertexID);
          master.addReplicaLocalID(peerIndex, remoteID);
        }
        step++;
        count = 0;
      }
    }
    Assert.assertEquals(step, registerSteps);

    //last step
    peer.sync();
    GraphJobMessage msg = null;
    long remoteID = 0;
    registerCount += peer.getNumCurrentMessages();
    while ((msg = peer.getCurrentMessage()) != null) {
      Assert.assertEquals(msg.isRegisterMessage(), true);
      
      int peerIndex = msg.getPeerIndex().get();
      V vertexID = (V) msg.getVertexId();
      remoteID = msg.getLocalID().get();
      
      Vertex<V, E, M> master = verticesMap.get(vertexID);
      master.addReplicaLocalID(peerIndex, remoteID);
    }
    
    replicasMap = null;
    LOG.info("Register " + registerCount
          + " replicas finished at " + peer.getSuperstepCount() + " steps.");
  }

  /**
   * Counts vertices globally by sending the count of vertices in the map to the
   * other peers.
   */
  private void countGlobal(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer)
      throws IOException, SyncException, InterruptedException {
    
    for (String peerName : peer.getAllPeerNames())
      peer.send(peerName, new GraphJobMessage(
          new IntWritable(vertices.size()), new IntWritable(replicas.length)));

    peer.sync();
    GraphJobMessage msg = null;
    while ((msg = peer.getCurrentMessage()) != null) {
      if (msg.isSizeMessage()) {
        numberVertices += msg.getVerticesSize().get();
        numberReplicas += msg.getReplicasSize().get();
      }
    }
    this.globalActiveVertexCount = this.numberVertices;  // TODO: selective

    if (isMasterTask(peer))
      peer.getCounter(GraphJobCounter.INPUT_VERTICES).increment(numberVertices);
    
    LOG.info("Counting finished at " + peer.getSuperstepCount() + " steps.");
    LOG.info("Total # of vertices: " + numberVertices + " and # of replicas: " + numberReplicas);
  }

  /**
   * @return the number of vertices, globally accumulated.
   */
  public final long getNumberVertices() {
    return numberVertices;
  }

  /**
   * @return the current number of iterations.
   */
  public final long getNumberIterations() {
    return iteration;
  }

  /**
   * @return the defined number of maximum iterations, -1 if not defined.
   */
  public final int getMaxIteration() {
    return maxIteration;
  }

  /**
   * @return the defined partitioner instance.
   */
  public final Partitioner<V, M> getPartitioner() {
    return partitioner;
  }

  /**
   * Gets the last aggregated value at the given index. The index is dependend
   * on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final Writable getLastAggregatedValue(int index) {
    return aggregationRunner.getLastAggregatedValue(index);
  }

  /**
   * Gets the last aggregated number of vertices at the given index. The index
   * is dependend on how the aggregators were configured during job setup phase.
   * 
   * @return the value of the aggregator, or null if none was defined.
   */
  public final IntWritable getNumLastAggregatedVertices(int index) {
    return aggregationRunner.getNumLastAggregatedVertices(index);
  }

  /**
   * @return the peer instance.
   */
  public final BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> getPeer() {
    return peer;
  }

  /**
   * Checks if this is a master task. The master task is the first peer in the
   * peer array.
   */
  public static boolean isMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName().equals(getMasterTask(peer));
  }

  /**
   * @return the name of the master peer, the name at the first index of the
   *         peers.
   */
  public static String getMasterTask(
      BSPPeer<Writable, Writable, Writable, Writable, GraphJobMessage> peer) {
    return peer.getPeerName(0);
  }

  /**
   * @return a new vertex instance
   */
  public static <V extends Writable, E extends Writable, M extends Writable> Vertex<V, E, M> newVertexInstance(
      Class<?> vertexClass, Configuration conf) {
    @SuppressWarnings("unchecked")
    Vertex<V, E, M> vertex = (Vertex<V, E, M>) ReflectionUtils.newInstance(
        vertexClass, conf);
    return vertex;
  }

}
